#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import get_url, INSTANCE_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, str_type_conv, is_python_2
from ibm_watson_machine_learning.metanames import ModelMetaNames, ExperimentMetaNames, FunctionMetaNames, PipelineMetanames, SpacesMetaNames, MemberMetaNames, FunctionNewMetaNames
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.models import Models
from ibm_watson_machine_learning.experiments import Experiments
from ibm_watson_machine_learning.functions import Functions
from ibm_watson_machine_learning.pipelines import Pipelines
from ibm_watson_machine_learning.spaces import Spaces
from multiprocessing import Pool
from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
from ibm_watson_machine_learning.href_definitions import API_VERSION, SPACES,PIPELINES, LIBRARIES, EXPERIMENTS, RUNTIMES, DEPLOYMENTS
import os
from warnings import warn
_DEFAULT_LIST_LENGTH = 50


class Repository(WMLResource):
    """
    Store and manage your models, functions, spaces, pipelines and experiments using Watson Machine Learning Repository.
    
    .. important::
    
        #. To view ModelMetaNames, use: \n 
           >>> client.repository.ModelMetaNames.show()
        #. To view ExperimentMetaNames, use: \n 
           >>> client.repository.ExperimentMetaNames.show()
        #. To view FunctionMetaNames, use: \n 
           >>> client.repository.FunctionMetaNames.show()
        #. To view PipelineMetaNames, use: \n 
           >>> client.repository.PipelineMetaNames.show()

    """

    cloud_platform_spaces = False
    icp_platform_spaces = False

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        if not client.ICP and not client.WSD and not client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            Repository._validate_type(client.service_instance.details, u'instance_details', dict, True)
            Repository._validate_type_of_details(client.service_instance.details, INSTANCE_DETAILS_TYPE)
        self._ICP = client.ICP
        self._WSD = client.WSD
        self._ml_repository_client = None
        Repository.cloud_platform_spaces = client.CLOUD_PLATFORM_SPACES
        Repository.icp_platform_spaces = client.ICP_PLATFORM_SPACES

        self.ExperimentMetaNames = ExperimentMetaNames()
        if not client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            self.FunctionMetaNames = FunctionMetaNames()
        else:
            self.FunctionMetaNames = FunctionNewMetaNames()
        self.PipelineMetaNames = PipelineMetanames()
        self.SpacesMetaNames = SpacesMetaNames()
        self.ModelMetaNames = ModelMetaNames()
        self.MemberMetaNames = MemberMetaNames()

        self._refresh_repo_client() # regular token is initialized in service_instance

    def _refresh_repo_client(self):
        # If apiKey is passed in credentials then refresh repoclient with IAM token else MLToken
        self._ml_repository_client = MLRepositoryClient(self._wml_credentials[u'url'])
        if self._client.proceed is True:
            if self._client.service_instance._is_iam() is not None:
                self._ml_repository_client.authorize_with_token(self._client.wml_token)
                self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                if self._client.project_id is not None:
                    self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)
            else:
                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    platform_spaces = True
                else:
                    platform_spaces = False

                self._ml_repository_client.authorize_with_iamtoken(self._client.wml_token,
                                                                   self._wml_credentials[u'instance_id'],
                                                                   platform_spaces)
                self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                # Cloud Convergence
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                   self._ml_repository_client._add_header('ML-Instance-ID', self._wml_credentials[u'instance_id'])
                if self._client.project_id is not None:
                    self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)
        else:
            if self._client._is_IAM():
                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    platform_spaces = True
                else:
                    platform_spaces = False

                self._ml_repository_client.authorize_with_iamtoken(self._client.wml_token,
                                                                   self._wml_credentials[u'instance_id'],
                                                                   platform_spaces)
                self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                # Cloud Convergence
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                   self._ml_repository_client._add_header('ML-Instance-ID', self._wml_credentials[u'instance_id'])
                if self._client.project_id is not None:
                    self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)
            else:
                if self._ICP:
                    self._repotoken = self._client._get_icptoken()
                    self._ml_repository_token = self._repotoken.replace('Bearer', '')
                    self._ml_repository_client.authorize_with_token(self._ml_repository_token)
                else:
                    if not self._client.WSD:
                        self._ml_repository_client.authorize(self._wml_credentials[u'username'], self._wml_credentials[u'password'])
                        self._ml_repository_client._add_header('X-WML-User-Client', 'PythonClient')
                        if self._client.project_id is not None:
                           self._ml_repository_client._add_header('X-Watson-Project-ID', self._client.project_id)

    def store_experiment(self, meta_props):
        """
        Create an experiment.

        **Parameters**

        .. important::
            #. **meta_props**:  meta data of the experiment configuration. To see available meta names use:\n
                >>> client.experiments.ConfigurationMetaNames.get()
               **type**: dict\n

        **Output**

        .. important::

            **returns**: Metadata of the experiment created\n
            **return type**: dict\n

        **Example**

        >>> metadata = {
        >>>  client.experiments.ConfigurationMetaNames.NAME: 'my_experiment',
        >>>  client.experiments.ConfigurationMetaNames.EVALUATION_METRICS: ['accuracy'],
        >>>  client.experiments.ConfigurationMetaNames.TRAINING_REFERENCES: [
        >>>      {
        >>>        'pipeline': {'href': pipeline_href_1}
        >>>      },
        >>>      {
        >>>        'pipeline': {'href':pipeline_href_2}
        >>>      },
        >>>   ]
        >>> }
        >>> experiment_details = client.repository.store_experiment(meta_props=metadata)
        >>> experiment_href = client.repository.get_experiment_href(experiment_details)
        """

        if self._client.WSD:
            raise WMLClientError(u'Experiment APIs are not supported in Watson Studio Desktop.')

        return self._client.experiments.store(meta_props)


    def store_space(self, meta_props):
        """
               Create a space.

               **Parameters**

               .. important::

                    #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
                                        >>> client.spaces.ConfigurationMetaNames.get()

                       **type**: dict\n

               **Output**

               .. important::

                    **returns**: Metadata of the space created\n
                    **return type**: dict\n

               **Example**

                >>> metadata = {
                >>>  client.spaces.ConfigurationMetaNames.NAME: 'my_space'
                >>> }
                >>> space_details = client.repository.store_space(meta_props=metadata)
                >>> space_href = client.repository.get_space_href(experiment_details)
        """


        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.store(meta_props)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_member(self, space_uid,meta_props):
        """
                Create a member within a space.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the member configuration. To see available meta names use:\n
                                    >>> client.spaces.ConfigurationMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the stored member\n
                    **return type**: dict\n

                .. note::
                    * client.spaces.MemberMetaNames.ROLE can be any one of the following "viewer, editor, admin"\n
                    * client.spaces.MemberMetaNames.IDENTITY_TYPE can be any one of the following "user,service"\n
                    * client.spaces.MemberMetaNames.IDENTITY can be either service-ID or IAM-userID\n

                **Example**

                 >>> metadata = {
                 >>>  client.spaces.MemberMetaNames.ROLE:"Admin",
                 >>>  client.spaces.MemberMetaNames.IDENTITY:"iam-ServiceId-5a216e59-6592-43b9-8669-625d341aca71",
                 >>>  client.spaces.MemberMetaNames.IDENTITY_TYPE:"service"
                 >>> }
                 >>> members_details = client.repository.create_member(space_uid=space_id, meta_props=metadata)
        """
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.create_member(space_uid,meta_props)

    @staticmethod
    def _meta_props_to_repository_v3_style(meta_props):
        if is_python_2():
            new_meta_props = meta_props.copy()

            for key in new_meta_props:
                if type(new_meta_props[key]) is unicode:
                    new_meta_props[key] = str(new_meta_props[key])

            return new_meta_props
        else:
            return meta_props

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store_pipeline(self, meta_props):
        """
                Create a pipeline.

                **Parameters**

                .. important::

                   #. **meta_props**:  meta data of the pipeline configuration. To see available meta names use:\n

                        >>> client.pipelines.ConfigurationMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Metadata of the pipeline created\\n
                    **return type**: dict\n

                **Example**

                    >>> metadata = {
                    >>>  client.pipelines.ConfigurationMetaNames.NAME: 'my_training_definition',
                    >>>  client.pipelines.ConfigurationMetaNames.DOCUMENT: {"doc_type":"pipeline","version": "2.0","primary_pipeline": "dlaas_only","pipelines": [{"id": "dlaas_only","runtime_ref": "hybrid","nodes": [{"id": "training","type": "model_node","op": "dl_train","runtime_ref": "DL","inputs": [],"outputs": [],"parameters": {"name": "tf-mnist","description": "Simple MNIST model implemented in TF","command": "python3 convolutional_network.py --trainImagesFile ${DATA_DIR}/train-images-idx3-ubyte.gz --trainLabelsFile ${DATA_DIR}/train-labels-idx1-ubyte.gz --testImagesFile ${DATA_DIR}/t10k-images-idx3-ubyte.gz --testLabelsFile ${DATA_DIR}/t10k-labels-idx1-ubyte.gz --learningRate 0.001 --trainingIters 6000","compute": {"name": "k80","nodes": 1},"training_lib_href":"/v4/libraries/64758251-bt01-4aa5-a7ay-72639e2ff4d2/content"},"target_bucket": "wml-dev-results"}]}]}}
                    >>> pipeline_details = client.repository.store_pipeline(pipeline_filepath, meta_props=metadata)
                    >>> pipeline_href = client.repository.get_pipeline_href(pipeline_details)
        """


        return self._client.pipelines.store(meta_props)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store_model(self, model, meta_props=None, training_data=None, training_target=None, pipeline=None, feature_names=None, label_column_names=None,subtrainingId=None, round_number=None):
        """
                Create a model.

                **Parameters**

                .. important::

                    #. **model**:  \n
                            Can be one of following:\n
                            - The train model object:\n
                                - scikit-learn
                                - xgboost
                                - spark (PipelineModel)
                            - path to saved model in format:\n
                                - keras (.tgz)
                                - pmml (.xml)
                                - scikit-learn (.tar.gz)
                                - tensorflow (.tar.gz)
                                - spss (.str)
                                - spark (.tar.gz)
                            - directory containing model file(s):\n
                                - scikit-learn
                                - xgboost
                                - tensorflow
                            - unique id of trained model

                    #. **Example**:
                    >>> published_model_details = wml_client.repository.store_model(model="rf_model_0.4.tar.gz", meta_props=metadata)
                    #. **training_data**:  Spark DataFrame supported for spark models. Pandas dataframe, numpy.ndarray or array supported for scikit-learn models\n
                       **type**: spark dataframe, pandas dataframe, numpy.ndarray or array\n

                    #. **meta_props**:  meta data of the models configuration. To see available meta names use:\n
                                        >>> client.repository.ModelMetaNames.get()

                       **type**: dict\n

                    #. **training_target**:  array with labels required for scikit-learn models\n
                       **type**: array\n

                    #. **pipeline**:  pipeline required for spark mllib models\n
                       **type**: object\n

                    #. **feature_names**:  Feature names for the training data in case of Scikit-Learn/XGBoost models. This is applicable only in the case where the training data is not of type - pandas.DataFrame.\n
                       **type**: numpy.ndarray or list\n

                    #. **label_column_names**:  Label column names of the trained Scikit-Learn/XGBoost models.\n
                       **type**: numpy.ndarray and list\n

                    #. **round_number**: round number of a Federated Learning experiment that has been configured to save intermediate models.  This applies when model is a training id.  
                       **type**: int



                **Output**

                .. important::

                    **returns**: Metadata of the model created\n
                    **return type**: dict\n

                .. note::

                    * For a keras model, model content is expected to contain a .h5 file and an archived version of it.\n

                    * feature_names is an optional argument containing the feature names for the training data in case of Scikit-Learn/XGBoost models. Valid types are numpy.ndarray and list. This is applicable only in the case where the training data is not of type - pandas.DataFrame.\n

                    * If the training data is of type pandas.DataFrame and feature_names are provided, feature_names are ignored.\n

                    * The value can be a single dictionary(being deprecated, use list even for single schema) or
                      a list if you are using single input data schema. you can provide multiple schemas as dictionaries inside a list.
                **Example**

                    >>> stored_model_details = client.repository.store_model(model, name)

                    In more complicated cases you should create proper metadata, similar to this one:\n

                    >>> sw_spec_id = client.software_specifications.get_id_by_name('scikit-learn_0.23-py3.7')
                    >>> sw_spec_id
                    >>> metadata = {
                    >>>        client.repository.ModelMetaNames.NAME: 'customer satisfaction prediction model',
                    >>>        client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_id,
                    >>>        client.repository.ModelMetaNames.TYPE: 'scikit-learn_0.23'
                    >>>}

                    In case when you want to provide input data schema of the model, you can provide it as part of meta

                    >>> sw_spec_id = client.software_specifications.get_id_by_name('spss-modeler_18.1')
                    >>> sw_spec_id
                    >>> metadata = {
                    >>>        client.repository.ModelMetaNames.NAME: 'customer satisfaction prediction model',
                    >>>        client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_id,
                    >>>        client.repository.ModelMetaNames.TYPE: 'spss-modeler_18.1',
                    >>>        client.repository.ModelMetaNames.INPUT_DATA_SCHEMA: [{'id': 'test',
                    >>>                                                             'type': 'list',
                    >>>                                                             'fields': [{'name': 'age', 'type': 'float'},
                    >>>                                                                        {'name': 'sex', 'type': 'float'},
                    >>>                                                                         {'name': 'fbs', 'type': 'float'},
                    >>>                                                                         {'name': 'restbp', 'type': 'float'}]
                    >>>                                                               },
                    >>>                                                               {'id': 'test2',
                    >>>                                                                'type': 'list',
                    >>>                                                                'fields': [{'name': 'age', 'type': 'float'},
                    >>>                                                                           {'name': 'sex', 'type': 'float'},
                    >>>                                                                           {'name': 'fbs', 'type': 'float'},
                    >>>                                                                           {'name': 'restbp', 'type': 'float'}]
                    >>>                                                               }]
                    >>>             }

                    A way you might use me with local tar.gz containing model:\n

                    >>> stored_model_details = client.repository.store_model(path_to_tar_gz, meta_props=metadata, training_data=None)

                    A way you might use me with local directory containing model file(s):\n

                    >>> stored_model_details = client.repository.store_model(path_to_model_directory, meta_props=metadata, training_data=None)

                    A way you might use me with trained model guid:\n

                    >>> stored_model_details = client.repository.store_model(trained_model_guid, meta_props=metadata, training_data=None)
            """

        return self._client._models.store(model, meta_props=meta_props, training_data=training_data, training_target=training_target, pipeline=pipeline, feature_names=feature_names, label_column_names=label_column_names,subtrainingId=subtrainingId,round_number=round_number)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def clone(self, artifact_id, space_id=None, action="copy", rev_id=None):
        # """ it is not supported in v4ga
        #     Creates a new resource(models, runtimes, libraries, experiments, functions, pipelines) identical with the model either in the same space or in a new space. All dependent assets will be cloned too.
        #
        #     **Parameters**
        #
        #     .. important::
        #         #. **model_id**:  Guid of the artifact to be cloned:\n
        #
        #            **type**: str\n
        #
        #         #. **space_id**: Guid of the space to which the model needs to be cloned. (optional)
        #
        #            **type**: str\n
        #
        #         #. **action**: Action specifying "copy" or "move". (optional)
        #
        #            **type**: str\n
        #
        #         #. **rev_id**: Revision ID of the artifact. (optional)
        #
        #            **type**: str\n
        #
        #     **Output**
        #
        #     .. important::
        #
        #             **returns**: Metadata of the model cloned.\n
        #             **return type**: dict\n
        #
        #     **Example**
        #
        #      >>> client.repository.clone(artifact_id=artifact_id,space_id=space_uid,action="copy")
        #
        #     .. note::
        #         * If revision id is not specified, all revisions of the artifact are cloned\n
        #
        #         * Default value of the parameter action is copy\n
        #
        #         * Space guid is mandatory for move action\n
        #
        #     """
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError('Cloning is not supported.')
        artifact = str_type_conv(artifact_id)
        Models._validate_type(artifact, 'artifact_id', STR_TYPE, True)
        space = str_type_conv(space_id)
        rev = str_type_conv(rev_id)
        action = str_type_conv(action)
        clone_meta = {}
        if space is not None:
            clone_meta["space"] = {"href": API_VERSION + SPACES + "/" + space}
        if action is not None:
            clone_meta["action"] = action
        if rev is not None:
            clone_meta["rev"] = rev
        res = self._check_artifact_type(artifact_id)

        url = ""
        type = ""
        if res['model'] is True:
            url = self._client.service_instance._href_definitions.get_published_model_href(artifact_id)
            type = "model"
        elif res['library'] is True:
            url = self._client.service_instance._href_definitions.get_custom_library_href(artifact_id)
            type = "library"
        elif res['runtime'] is True:
            url = self._client.service_instance._href_definitions.get_runtime_href(artifact_id)
            type = "runtime"
        elif res['function'] is True:
            url = self._client.service_instance._href_definitions.get_function_href(artifact_id)
            type = "function"
        elif res['pipeline'] is True:
            url = self._client.service_instance._href_definitions.get_pipeline_href(artifact_id)
            type = "pipeline"
        elif res['experiment'] is True:
            url = self._client.service_instance._href_definitions.get_experiment_href(artifact_id)
            type = "experiment"


        if type == "":
            raise WMLClientError('Unsupported artifact type. Supported artifact types are models, libraries, runtimes, experiments, pipelines and functions')

        response_post = requests.post(url, json=clone_meta,
                                      headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=200, operationName=u'cloning '+ type,
                                        response=response_post)

        return details

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store_function(self, function, meta_props):
        """
            Create a function.

            **Parameters**

            .. important::
                #. **meta_props**:  meta data or name of the function. To see available meta names use:\n
                                    >>> client.repository.FunctionMetaNames.show()

                       **type**: dict\n
                #. **function**:  path to file with archived function content or function (as described above)\n
                         - As a 'function' may be used one of the following:\n
                         - filepath to gz file\n
                         - 'score' function reference, where the function is the function which will be deployed\n
                         - generator function, which takes no argument or arguments which all have primitive python default values and as result return 'score' function\n
                   **type**: str or function\n

            **Output**

            .. important::

                    **returns**: Metadata of the function created.\n
                    **return type**: dict\n

            **Example**

                 The most simple use is (using `score` function):\n

                 >>> meta_props = {
                 >>>    client.repository.FunctionMetaNames.NAME: "function",
                 >>>    client.repository.FunctionMetaNames.DESCRIPTION: "This is ai function",
                 >>>    client.repository.FunctionMetaNames.SOFTWARE_SPEC_UID: "53dc4cf1-252f-424b-b52d-5cdd9814987f"}

                 >>> def score(payload):
                 >>>      values = [[row[0]*row[1]] for row in payload['values']]
                 >>>      return {'fields': ['multiplication'], 'values': values}
                 >>> stored_function_details = client.repository.store_function(score, meta_props)

                 Other, more interesting example is using generator function.
                 In this situation it is possible to pass some variables:

                    >>> wml_creds = {...}
                    >>> def gen_function(wml_credentials=wml_creds, x=2):
                    >>>        def f(payload):
                    >>>            values = [[row[0]*row[1]*x] for row in payload['values']]
                    >>>            return {'fields': ['multiplication'], 'values': values}
                    >>>        return f
                    >>> stored_function_details = client.repository.store_function(gen_function, meta_props)
            """

        return self._client._functions.store(function, meta_props)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_model_revision(self, model_uid):
        """
        Create a new version for a model.

        **Parameters**

        .. important::

            #. **model_uid**:  Model ID.\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: Model version details.\n
            **return type**: dict\n

        **Example**

            >>> stored_model_revision_details = client.repository.create_model_revision( model_uid="MODELID")
        """

        return self._client._models.create_revision(model_uid=model_uid)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_pipeline_revision(self, pipeline_uid):
        """
        Create a new version for a model.

        **Parameters**

        .. important::

            #. **pipeline_uid**:  Unique ID of the Pipeline.\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: Pipeline version details.\n
            **return type**: dict\n

        **Example**

            >>> stored_pipeline_revision_details = client.repository.create_pipeline_revision( pipeline_uid)
        """

        return self._client.pipelines.create_revision(pipeline_uid=pipeline_uid)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_function_revision(self, function_uid):
        """
        Create a new version for a function.

        **Parameters**

        .. important::

            #. **function_uid**:  Unique ID of the function.\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: Function version details.\n
            **return type**: dict\n

        **Example**

            >>> stored_function_revision_details = client.repository.create_function_revision( function_uid)
        """

        return self._client._functions.create_revision(function_uid=function_uid)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_experiment_revision(self, experiment_uid):
        """
        Create a new version for a experiment.

        **Parameters**

        .. important::

            #. **experiment_uid**:  Unique ID of the experiment.\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: experiment version details.\n
            **return type**: dict\n

        **Example**

            >>> stored_experiment_revision_details = client.repository.create_experiment_revision(experiment_uid)
        """

        return self._client.experiments.create_revision(experiment_id=experiment_uid)

    def update_model(self, model_uid, updated_meta_props=None, update_model=None):
        """
        Updates existing model metadata.

        **Parameters**

        .. important::

            #. **model_uid**: Unique id of model which definition should be updated\n
               **type**: str\n

            #. **updated_meta_props**: elements which should be changed, where keys are ConfigurationMetaNames\n
               **type**: dict\n

            #. **update_model**: archived model content file or path to directory containing archived model file which should be changed for specific model_uid.
               This parameters is valid only for CP4D 3.0.0.\n
               **type**: object or archived model content file\n

        **Output**

        .. important::

            **returns**: metadata of updated model\n
            **return type**: dict\n

        **Example 1**

         >>> metadata = {
         >>> client.repository.ModelMetaNames.NAME:"updated_model"
         >>> }
         >>> model_details = client.repository.update_model(model_uid, updated_meta_props=metadata)

        **Example 2**

         >>> metadata = {
         >>> client.repository.ModelMetaNames.NAME:"updated_model"
         >>> }
         >>> model_details = client.repository.update_model(model_uid, updated_meta_props=metadata, update_model="newmodel_content.tar.gz")
        """

        return self._client._models.update(model_uid, updated_meta_props, update_model)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update_experiment(self, experiment_uid, changes):
        """
                Updates existing experiment metadata.

                **Parameters**

                .. important::

                    #. **experiment_uid**: Unique of Id experiment which definition should be updated\n
                       **type**: str\n
                    #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                       **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of updated experiment\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>> client.repository.ExperimentMetaNames.NAME:"updated_exp"
                 >>> }
                 >>> exp_details = client.repository.update_experiment(experiment_uid, changes=metadata)

        """
        if self._client.WSD:
            raise WMLClientError('Experiments APIs are not supported in IBM Watson Studio Desktop.')

        return self._client.experiments.update(experiment_uid, changes)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update_function(self, function_uid, changes, update_function=None):
        """
        Updates existing function metadata.

        **Parameters**

        .. important::

            #. **function_uid**: Unique Id of function which define what should be updated\n
               **type**: str\n

            #. **changes**: Elements which should be changed, where keys are ConfigurationMetaNames.\n
               **type**: dict\n

            #. **update_function**:  Path to file with archived function content or function which should be changed for specific function_uid.
               This parameters is valid only for CP4D 3.0.0.\n
               **type**: str or function\n

        **Output**

        .. important::

            **returns**: metadata of updated function\n
            **return type**: dict\n

        **Example**

            >>> metadata = {
            >>> client.repository.FunctionMetaNames.NAME:"updated_function"
            >>> }
            >>>
            >>> function_details = client.repository.update_function(function_uid, changes=metadata)
        """
        return self._client._functions.update(function_uid, changes, update_function)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update_pipeline(self, pipeline_uid, changes):
        """
                Updates existing pipeline metadata.

                **Parameters**

                .. important::

                    #. **pipeline_uid**:  Unique Id of pipeline which definition should be updated\n
                       **type**: str\n
                    #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                       **type**: dict\n

                **Output**

                .. important::

                   **returns**: metadata of updated pipeline\n
                   **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>> client.repository.PipelineMetanames.NAME:"updated_pipeline"
                 >>> }
                 >>> pipeline_details = client.repository.update_pipeline(pipeline_uid, changes=metadata)

        """

        return self._client.pipelines.update(pipeline_uid, changes)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update_space(self, space_uid, changes):
        """
                Updates existing space metadata.

                **Parameters**

                .. important::
                    #. **space_uid**:  Unique Id of space which definition should be updated\n
                       **type**: str\n
                    #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                       **type**: dict\n

                **Output**

                .. important::
                    **returns**: metadata of updated space\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>> client.repository.SpacesMetaNames.NAME:"updated_space"
                 >>> }
                 >>> space_details = client.repository.update_space(space_uid, changes=metadata)
        """
        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.update(space_uid, changes)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def load(self, artifact_uid):
        """
           Load model from repository to object in local environment.

           **Parameters**

           .. important::

                #. **artifact_uid**:  Unique Id of model\n
                   **type**: str\n

           **Output**

           .. important::

                **returns**: model object\n
                **return type**: object\n

           **Example**

            >>> model_obj = client.repository.load(model_uid)
        """

        return self._client._models.load(artifact_uid)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def download(self, artifact_uid, filename='downloaded_artifact.tar.gz', rev_uid=None, format=None):
        """
               Downloads configuration file for artifact with specified uid.

               **Parameters**

               .. important::

                    #. **artifact_uid**:  Unique Id of model, function, runtime or library\n
                       **type**: str\n

                    #. **filename**:  Name of the file to which the artifact content has to be downloaded\n
                       **default value**: downloaded_artifact.tar.gz\n
                       **type**: str\n

               **Output**

               .. important::

                   **returns**: Path to the downloaded artifact content\n
                   **return type**: str\n

               .. note::

                    If filename is not specified, the default filename is "downloaded_artifact.tar.gz".\n

               **Example**

                >>> client.repository.download(model_uid, 'my_model.tar.gz')
                >>> client.repository.download(model_uid, 'my_model.json') # if original model was saved as json, works only for xgboost 1.3
        """
        self._validate_type(artifact_uid, 'artifact_uid', STR_TYPE, True)
        self._validate_type(filename, 'filename', STR_TYPE, True)

        res = self._check_artifact_type(artifact_uid)

        if res['model'] is True:
            return self._client._models.download(artifact_uid, filename, rev_uid,format)
        elif res['function'] is True:
            return self._client._functions.download(artifact_uid, filename, rev_uid)
        elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_35 and not self._client.ICP_40  \
                and not self._client.ICP_45 and not self._client.ICP_46 and res['library'] is True:
            return self._client.runtimes.download_library(artifact_uid, filename)
        elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_35 and not self._client.ICP_40 \
                and not self._client.ICP_45 and not self._client.ICP_46 and res['runtime'] is True:
            return self._client.runtimes.download_configuration(artifact_uid, filename)
        else:
            raise WMLClientError('Unexpected type of artifact to download or Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid) )

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, artifact_uid):
        """
           Delete model, experiment, pipeline, space, runtime, library or function from repository.

           **Parameters**

           .. important::
                #. **artifact_uid**:  Unique id of stored model, experiment, function, pipeline, space, library or runtime \n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

           **Example**

            >>> client.repository.delete(artifact_uid)
        """

        artifact_uid = str_type_conv(artifact_uid)
        Repository._validate_type(artifact_uid, u'artifact_uid', STR_TYPE, True)
        if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and self._if_deployment_exist_for_asset(artifact_uid):
            raise WMLClientError(
                u'Cannot delete artifact that has existing deployments. Please delete all associated deployments and try again')
        params = self._client._params()
        if Repository.cloud_platform_spaces or self._client.ICP_PLATFORM_SPACES:
            # ideally purge_on_delete=true query param has to be provided for deletion of cams assets
            # This doesn't seem to be done for CP4D 3.0.1 and before. We should do this for CP4D 3.5
            params.update({'purge_on_delete': 'true'})

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(artifact_uid),
                                       params=params,
                                       headers=self._client._get_headers())

        if response.status_code == 200 or response.status_code == 204:
            if response.status_code == 200:
                response = self._handle_response(200, u'delete assets', response)
                return response
            else:
                response = self._handle_response(204, u'delete assets', response)
                return response
        else:
            if Repository.cloud_platform_spaces or self._client.ICP_PLATFORM_SPACES:
                # Since we are using /v2/assets for deletion, don't need all the logic
                # in the following else block. The else block is applicable only for cloud beta
                # and has to be kept till then. For 3.5, move logic to same as cloud convergence
                # for deletion
                if response.status_code == 404:
                    raise WMLClientError(u'Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid))
                else:
                    raise WMLClientError("Deletion error for the given id : ", response.text)
            else:
                artifact_type = self._check_artifact_type(artifact_uid)
                self._logger.debug(u'Attempting deletion of artifact with type: \'{}\''.format(str(artifact_type)))
                if self._client.WSD:
                    if artifact_type[u'model'] is True:
                        return self._client._models.delete(artifact_uid)
                    elif artifact_type[u'pipeline'] is True:
                        return self._client.pipelines.delete(artifact_uid)
                    elif artifact_type[u'function'] is True:
                        return self._client._functions.delete(artifact_uid)
                    else:
                        raise WMLClientError(u'Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid))
                else:
                    if artifact_type[u'model'] is True:
                        return self._client._models.delete(artifact_uid)
                    elif artifact_type[u'experiment'] is True:
                        return self._client.experiments.delete(artifact_uid)
                    elif artifact_type[u'pipeline'] is True:
                        return self._client.pipelines.delete(artifact_uid)
                    elif artifact_type[u'function'] is True:
                        return self._client._functions.delete(artifact_uid)
                    elif artifact_type[u'space'] is True:
                        return self._client.spaces.delete(artifact_uid)
                    elif artifact_type[u'runtime'] is True:
                        return self._client.runtimes.delete(artifact_uid)
                    elif artifact_type[u'library'] is True:
                        return self._client.runtimes.delete_library(artifact_uid)
                    else:
                        raise WMLClientError(u'Artifact with artifact_uid: \'{}\' does not exist.'.format(artifact_uid))

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, artifact_uid=None):
        """
           Get metadata of stored artifacts. If artifact_uid is not specified returns all models, experiments, functions, pipelines, spaces, libraries and runtimes metadata.

           **Parameters**

           .. important::
                #. **artifact_uid**: Unique Id of stored model, experiment, function, pipeline, space, library or runtime (optional)\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: stored artifact(s) metadata\n
                **return type**: dict\n
                dict (if artifact_uid is not None) or {"resources": [dict]} (if artifact_uid is None)\n

           .. note::
                If artifact_uid is not specified, all models, experiments, functions, pipelines, spaces, libraries and runtimes metadata is fetched\n

           **Example**

            >>> details = client.repository.get_details(artifact_uid)
            >>> details = client.repository.get_details()
        """

        artifact_uid = str_type_conv(artifact_uid)
        Repository._validate_type(artifact_uid, u'artifact_uid', STR_TYPE, False)

        if artifact_uid is None and self._client.WSD is None:
                model_details = self._client._models.get_details()
                experiment_details = self.get_experiment_details()
                pipeline_details = self.get_pipeline_details()
                function_details = self._client._functions.get_details()

                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                    space_details = self._client.spaces.get_details()
                    library_details = self._client.runtimes.get_library_details()
                    runtime_details = self._client.runtimes.get_details()
                    details = {
                        u'models': model_details,
                        u'experiments': experiment_details,
                        u'pipeline': pipeline_details,
                        u'runtimes': runtime_details,
                        u'libraries': library_details,
                        u'spaces': space_details,
                        u'functions': function_details
                    }
                else:
                    details = {
                        u'models': model_details,
                        u'experiments': experiment_details,
                        u'pipeline': pipeline_details,
                        u'functions': function_details
                    }
        else:
            if self._client.WSD and artifact_uid is None:
                raise WMLClientError(
                        u' artifiact_uid is mandatory for get_details() in IBM Watson Studio Desktop.')
            uid_type = self._check_artifact_type(artifact_uid)
            if self._client.WSD:
                if uid_type[u'model'] is True:
                    details = self._client._models.get_details(artifact_uid)
                elif uid_type[u'pipeline'] is True:
                    details = self.get_pipeline_details(artifact_uid)
                elif uid_type[u'function'] is True:
                    details = self._client._functions.get_details(artifact_uid)
                else:
                    raise WMLClientError(
                        u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))
            else:
                if uid_type[u'model'] is True:
                    details = self._client._models.get_details(artifact_uid)
                elif uid_type[u'experiment'] is True:
                    details = self.get_experiment_details(artifact_uid)
                elif uid_type[u'pipeline'] is True:
                    details = self.get_pipeline_details(artifact_uid)
                elif uid_type[u'function'] is True:
                    details = self._client._functions.get_details(artifact_uid)
                elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES and uid_type[u'runtime'] is True:
                    details = self._client.runtimes.get_details(artifact_uid)
                elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES and uid_type[u'library'] is True:
                    details = self._client.runtimes.get_library_details(artifact_uid)
                elif not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES and uid_type[u'space'] is True:
                    details = self._client.spaces.get_details(artifact_uid)
                else:
                    raise WMLClientError(u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))

        return details

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_model_details(self, model_uid=None, limit=None, asynchronous=False, get_all=False):
        """
           Get metadata of stored model. If model_uid is not specified returns all models metadata.

           **Parameters**

           .. important::
                #. **model_uid**: Unique Id of Model (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n
                #. **asynchronous**:  if True, it will work as a generator (optional)\n
                   **type**: bool\n
                #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
                   **type**: bool\n

           **Output**

           .. important::
                **returns**: metadata of model(s)\n
                **return type**: dict (if model_uid is not None) or {"resources": [dict]} (if model_uid is None)\n

           .. note::
                If model_uid is not specified, all models metadata is fetched\n

           **Example**

            >>> model_details = client.repository.get_model_details(model_uid)
            >>> models_details = client.repository.get_model_details()
            >>> models_details = client.repository.get_model_details(limit=100)
            >>> models_details = client.repository.get_model_details(limit=100, get_all=True)
            >>> models_details = []
            >>> for details in client.repository.get_model_details(limit=100, asynchronous=True, get_all=True):
            >>>     models_details.append(details)
        """

        return self._client._models.get_details(model_uid, limit,
                                                asynchronous=asynchronous, get_all=get_all)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_model_revision_details(self, model_uid, rev_uid):
        """
           Get metadata of model revision. 

           **Parameters**

           .. important::
                #. **experiment_uid**:  Unique Id of model\n
                   **type**: str\n
                #. **limit**:  Unique id of model revision\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: model revision metadata\n
                **return type**: dict\n

           **Example**

                 >>> model_rev_details = client.respository.get_model_revision_details(model_uid, rev_uid)

        """

        if not self._client.ICP_30 and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError('Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')
        return self._client._models.get_revision_details(model_uid, rev_uid)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_experiment_details(self, experiment_uid=None, limit=None):
        """
           Get metadata of experiment. If no experiment_uid is specified all experiments metadata is returned.

           **Parameters**

           .. important::
                #. **experiment_uid**: Unique Id of experiment (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: experiment(s) metadata\n
                **return type**: dict\n
                dict (if experiment_uid is not None) or {"resources": [dict]} (if experiment_uid is None)\n

           .. note::
                If experiment_uid is not specified, all experiments metadata is fetched\n

           **Example**

                 >>> experiment_details = client.respository.get_experiment_details(experiment_uid)

         """

        if self._client.WSD:
            raise WMLClientError('Experiment APIs are not supported in IBM Watson Studio Desktop.')

        experiment_uid = str_type_conv(experiment_uid)
        Repository._validate_type(experiment_uid, u'experiment_uid', STR_TYPE, False)
        Repository._validate_type(limit, u'limit', int, False)

        url = self._client.service_instance._href_definitions.get_experiments_href()

        return self._client.experiments.get_details(experiment_uid, limit)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_experiment_revision_details(self, experiment_uid, rev_id):
        """
           Get metadata of experiment revision. 

           **Parameters**

           .. important::
                #. **experiment_uid**: Unique Id of experiment\n
                   **type**: str\n
                #. **rev_id**:  Unique id of experiment revision\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: experiment revision metadata\n
                **return type**: dict\n

           **Example**

                 >>> experiment_rev_details = client.respository.get_experiment__revision_details(experiment_uid, rev_uid)

        """

        if not self._client.ICP_30 and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                'Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')

        return self._client.experiments.get_revision_details(experiment_uid, rev_id)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_function_details(self, function_uid=None, limit=None):
        """

           Get metadata of function. If no function_uid is specified all functions metadata is returned.

           **Parameters**

           .. important::
                #. **function_uid**:  Unique Id of function (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: function(s) metadata\n
                **return type**: dict (if function_uid is not None) or {"resources": [dict]} (if function_uid is None)\n

           .. note::
                If function_uid is not specified, all functions metadata is fetched\n

           **Example**

                >>> function_details = client.respository.get_function_details(function_uid)
                >>> function_details = client.respository.get_function_details()
         """
        Repository._validate_type(function_uid, u'function_uid', STR_TYPE, False)
        Repository._validate_type(limit, u'limit', int, False)
        return self._client._functions.get_details(function_uid, limit)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_function_revision_details(self, function_uid, rev_id):
        """

           Get metadata of function revision.

           **Parameters**

           .. important::
                #. **function_uid**:  Unique Id of function\n
                   **type**: str\n
                #. **rev_id**:  Unique Id of function revision\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: function revision metadata\n
                **return type**: dict\n

           **Example**

                >>> function_rev_details = client.respository.get_function_revision_details(function_uid, rev_id)

        """

        if not self._client.ICP_30 and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError('Not supported in this release')
        return self._client._functions.get_revision_details(function_uid, rev_id)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_pipeline_details(self, pipeline_uid=None, limit=None):
        """
           Get metadata of stored pipelines. If pipeline_uid is not specified returns all pipelines metadata.

           **Parameters**

           .. important::
                #. **pipeline_uid**: Unique id of Pipeline(optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of pipeline(s)\n
                **return type**: dict (if pipeline_uid is not None) or {"resources": [dict]} (if pipeline_uid is None)\n

           .. note::
                If pipeline_uid is not specified, all pipelines metadata is fetched\n

           **Example**

                >>> pipeline_details = client.repository.get_pipeline_details(pipeline_uid)
                >>> pipeline_details = client.repository.get_pipeline_details()
        """

        Repository._validate_type(pipeline_uid, u'pipeline_uid', STR_TYPE, False)
        Repository._validate_type(limit, u'limit', int, False)
        return self._client.pipelines.get_details(pipeline_uid, limit)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_pipeline_revision_details(self, pipeline_uid, rev_id):
        """
           Get metadata of stored pipeline revision. 

           **Parameters**

           .. important::
                #. **pipeline_uid**: Unique id of Pipeline\n
                   **type**: str\n
                #. **rev_id**:  Unique id Pipeline revision\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: metadata of revision pipeline(s)\n
                **return type**: dict\n

           **Example**

                >>> pipeline_rev_details = client.repository.get_pipeline_revision_details(pipeline_uid, rev_id)
        """

        if not self._client.ICP_30 and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                'Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')
        return self._client.pipelines.get_revision_details(pipeline_uid, rev_id)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_space_details(self, space_uid=None, limit=None):
        """
           Get metadata of stored space. If space_uid is not specified returns all model spaces metadata.

           **Parameters**

           .. important::
                #. **space_uid**: Unique id of Space (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of space(s)\n
                **return type**: dict (if space_uid is not None) or {"resources": [dict]} (if space_uid is None)\n

           .. note::
                If space_uid is not specified, all spaces metadata is fetched\n

           **Example**

            >>> space_details = client.repository.get_space_details(space_uid)
            >>> space_details = client.repository.get_space_details()
        """

        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        Repository._validate_type(space_uid, u'space_uid', STR_TYPE, False)
        Repository._validate_type(limit, u'limit', int, False)
        return self._client.spaces.get_details(space_uid, limit)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_members_details(self, space_uid, member_id=None, limit=None):
        """
           Get metadata of members associated with a space. If member_uid is not specified, it returns all the members metadata.

           **Parameters**

           .. important::
                #. **space_uid**: Unique id of member (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of member(s) of a space\n
                **return type**: dict (if member_id is not None) or {"resources": [dict]} (if member_id is None)\n

           .. note::
                If member id is not specified, all members metadata is fetched\n

           **Example**

            >>> member_details = client.repository.get_member_details(space_uid,member_id)
        """

        if self._client.WSD or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")
        return self._client.spaces.get_members_details(space_uid,member_id)


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_model_href(model_details):
        """
            Get href of stored model.

           **Parameters**

           .. important::
                #. **model_details**:  Metadata of the stored model\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of stored model\n
                **return type**: str\n

           **Example**

            >>> model_details = client.repository.get_model_detailsf(model_uid)
            >>> model_uid = client.repository.get_model_href(model_details)
        """

        return Models.get_href(model_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_model_uid(model_details):
        """
            This method is deprecated, please use get_id() instead."
        """
        warn("This method is deprecated, please use get_model_id()")
        print("This method is deprecated, please use get_model_id()")

        return Models.get_id(model_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_model_id(model_details):
        """
            Get Unique Id of stored model.

           **Parameters**

           .. important::
                #. **model_details**:  Metadata of the stored model\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: Unique Id of stored model\n
                **return type**: str\n

           **Example**

            >>> model_id = client.repository.get_model_id(model_details)
        """

        return Models.get_id(model_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_experiment_uid(experiment_details):
        """
           Get Unique Id of stored experiment.

           **Parameters**

           .. important::
                #. **experiment_details**:  Metadata of the stored experiment\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: Unique Id of stored experiment\n
                **return type**: str\n

           **Example**

            >>> experiment_details = client.repository.get_experiment_detailsf(experiment_uid)
            >>> experiment_uid = client.repository.get_experiment_uid(experiment_details)

        """

        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        return Experiments.get_uid(experiment_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_experiment_id(experiment_details):
        """
           Get Unique Id of stored experiment.

           **Parameters**

           .. important::
                #. **experiment_details**:  Metadata of the stored experiment\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: Unique Id of stored experiment\n
                **return type**: str\n

           **Example**

            >>> experiment_details = client.repository.get_experiment_details(experiment_uid)
            >>> experiment_uid = client.repository.get_experiment_id(experiment_details)

        """

        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        return Experiments.get_id(experiment_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_experiment_href(experiment_details):
        """
           Get href of stored experiment.

           **Parameters**

           .. important::
                #. **experiment_details**:  Metadata of the stored experiment\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of stored experiment\n
                **return type**: str\n

           **Example**

             >>> experiment_details = client.repository.get_experiment_detailsf(experiment_uid)
             >>> experiment_href = client.repository.get_experiment_href(experiment_details)

        """
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        return Experiments.get_href(experiment_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_function_id(function_details):
        """
            Get Id of stored function.

            **Parameters**

            .. important::
                #. **function_details**:  Metadata of the stored function\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Id of stored function\n
                **return type**: str\n

            **Example**

             >>> function_details = client.repository.get_function_details(function_uid)
             >>> function_id = client.repository.get_function_id(function_details)
        """
        return Functions.get_id(function_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_function_uid(function_details):
        """
            Get Unique Id of stored function. Deprecated!! Use get_function_id(function_details) instead

            **Parameters**

            .. important::
                #. **function_details**:  Metadata of the stored function\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Unique Id of stored function\n
                **return type**: str\n

            **Example**

             >>> function_details = client.repository.get_function_detailsf(function_uid)
             >>> function_uid = client.repository.get_function_uid(function_details)
        """
        return Functions.get_uid(function_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_pipeline_uid(pipeline_details):
        """
            Get pipeline_uid from pipeline details.

            **Parameters**

            .. important::
                #. **pipeline_details**:  Metadata of the stored pipeline\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Unique Id of pipeline\n
                **return type**: str

            **Example**

             >>> pipeline_details = client.repository.get_pipeline_details(pipeline_uid)
             >>> pipeline_uid = client.repository.get_pipeline_uid(pipeline_details)

        """

        return Pipelines.get_uid(pipeline_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_function_href(function_details):
        """

            Get href of stored function.

            **Parameters**

            .. important::
                #. **function_details**:  Metadata of the stored function\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: href of stored function\n
                **return type**: str\n

            **Example**

             >>> function_details = client.repository.get_function_detailsf(function_uid)
             >>> function_url = client.repository.get_function_href(function_details)
        """
        return Functions.get_href(function_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_pipeline_href(pipeline_details):
        """
            Get pipeline_hef from pipeline details.

            **Parameters**

            .. important::
                #. **pipeline_details**:  Metadata of the stored pipeline\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: pipeline href\n
                **return type**: str

            **Example**

             >>> pipeline_details = client.repository.get_pipeline_details(pipeline_uid)
             >>> pipeline_href = client.repository.get_pipeline_href(pipeline_details)
        """

        return Pipelines.get_href(pipeline_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_pipeline_id(pipeline_details):
        """
            Get pipeline_uid from pipeline details.

            **Parameters**

            .. important::
                #. **pipeline_details**:  Metadata of the stored pipeline\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Unique Id of pipeline\n
                **return type**: str

            **Example**

             >>> pipeline_details = client.repository.get_pipeline_details(pipeline_uid)
             >>> pipeline_uid = client.repository.get_pipeline_id(pipeline_details)

        """

        return Pipelines.get_id(pipeline_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_space_uid(space_details):
        """
            Get space_uid from space details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored space\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Unique Id of space\n
                **return type**: str

            **Example**

             >>> space_details = client.repository.get_space_details(space_uid)
             >>> space_uid = client.repository.get_space_uid(space_details)
        """
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_uid(space_details)

    @staticmethod
    def get_member_uid(member_details):
        """
            Get member_uid from member details.

            **Parameters**

            .. important::
                #. **member_details**:  Metadata of the created member\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: unique id of member\n
                **return type**: str

            **Example**

             >>> member_details = client.repository.get_member_details(member_id)
             >>> member_id = client.repository.get_member_uid(member_details)
        """
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_member_uid(member_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_space_href(space_details):
        """
            Get space_href from space details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored space\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: space href\n
                **return type**: str

            **Example**

             >>> space_details = client.repository.get_space_details(space_uid)
             >>> space_href = client.repository.get_space_href(space_details)
        """
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_href(space_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_member_href(member_details):
        """
            Get member_href from member details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored member\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: member href\n
                **return type**: str

            **Example**

             >>> member_details = client.repository.get_member_details(member_id)
             >>> member_href = client.repository.get_member_href(member_details)
        """
        if 'WSD_PLATFORM' in os.environ and os.environ['WSD_PLATFORM'] == 'True':
            raise WMLClientError(u'Spaces APIs are not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        return Spaces.get_member_href(member_details)

    def list(self):
        """
           List stored models, pipelines, runtimes, libraries, functions, spaces and experiments. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all models, pipelines, runtimes, libraries, functions, spaces and experiments in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list()
        """

        from tabulate import tabulate

        headers = self._client._get_headers()
        params = self._client._params()
        params.update({u'limit': 1000})
        #params = {u'limit': 1000} # TODO - should be unlimited, if results not sorted

        pool = Pool(processes=4)
        isIcp = self._ICP
        if self._client.WSD:
            raise WMLClientError(
                u'list() - Listing all artifact is not supported for IBM Watson Studio Desktop. '
                u'Use list method of specific artifact.')


        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            endpoints = {
                u'model': self._client.service_instance._href_definitions.get_published_models_href(),
                u'experiment': self._client.service_instance._href_definitions.get_experiments_href(),
                u'pipeline': self._client.service_instance._href_definitions.get_pipelines_href(),
                u'function': self._client.service_instance._href_definitions.get_functions_href()
            }
        else:
            endpoints = {
                u'model': self._client.service_instance._href_definitions.get_published_models_href(),
                u'experiment': self._client.service_instance._href_definitions.get_experiments_href(),
                u'pipeline': self._client.service_instance._href_definitions.get_pipelines_href(),
                u'function': self._client.service_instance._href_definitions.get_functions_href(),
                u'runtime': self._client.service_instance._href_definitions.get_runtimes_href(),
                u'library': self._client.service_instance._href_definitions.get_custom_libraries_href()
            }

        artifact_get = {}
        for artifact in endpoints:
            if (artifact=="library" or artifact=="runtime" or artifact=="space"):
                params = None
            else:
                params = self._client._params()
            artifact_get[artifact] = pool.apply_async(get_url,
                                                (endpoints[artifact], self._client._get_headers(), params, isIcp))

        # artifact_get = {artifact: pool.apply_async(get_url, (endpoints[artifact], headers, self._client._params(), isIcp)) for
        #           artifact in endpoints if (artifact != "library" or artifact != "runtime" or artifact != "space")}
        # artifact_no_space = {artifact: pool.apply_async(get_url, (endpoints[artifact], headers, None, isIcp)) for artifact
        #                    in endpoints if (artifact == "library" or artifact == "runtime")}
        # artifact_get.update(artifact_no_space)

        resources = {artifact: [] for artifact in endpoints}

        for artifact in endpoints:
            try:
                    response = artifact_get[artifact].get()
                    response_text = self._handle_response(200, u'getting all {}s'.format(artifact), response)
                    resources[artifact] = response_text[u'resources']
            except Exception as e:
                    self._logger.error(e)

        pool.close()

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            model_values = [(m[u'metadata'][u'id'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at'],
                             m[u'entity'][u'type'], u'model') for m in resources[u'model']]
            experiment_values = [
                (m[u'metadata'][u'id'], m[u'metadata'][u'name'], m['metadata']['created_at'], u'-', u'experiment') for m
                in resources[u'experiment']]
            pipeline_values = [
                (m[u'metadata'][u'id'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at'], u'-', u'pipeline') for m
                in self._client.pipelines.get_details()[u'resources']]
            function_values = [(m[u'metadata'][u'id'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at'], u'-',
                                m[u'entity'][u'type'] + u' function') for m in resources[u'function']]
            values = list(set(model_values + experiment_values + pipeline_values + function_values))
        else:
            model_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], m[u'entity'][u'type'], u'model') for m in resources[u'model']]
            experiment_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'],m['metadata']['created_at'], u'-', u'experiment') for m in resources[u'experiment']]
            pipeline_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', u'pipeline')for m in self._client.pipelines.get_details()[u'resources']]
            function_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', m[u'entity'][u'type'] + u' function') for m in resources[u'function']]
            runtime_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', m[u'entity'][u'platform'][u'name'] + u' runtime') for m in resources[u'runtime']]
            library_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at'], u'-', m[u'entity'][u'platform'][u'name'] + u' library') for m in resources[u'library']]
            values = list(set(model_values + experiment_values + pipeline_values + function_values + runtime_values + library_values))
        values = sorted(sorted(values, key=lambda x: x[2], reverse=True), key=lambda x: x[4])

        table = tabulate([[u'GUID', u'NAME', u'CREATED', u'FRAMEWORK', u'TYPE']] + values[:_DEFAULT_LIST_LENGTH])
        print(table)
        if len(values) > _DEFAULT_LIST_LENGTH:
             print('Note: Only first {} records were displayed. To display more use more specific list functions.'.format(_DEFAULT_LIST_LENGTH))


    def list_models(self, limit=None, asynchronous=False, get_all=False):
        """
           List stored models. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n
                #. **asynchronous**:  if True, it will work as a generator (optional)\n
                   **type**: bool\n
                #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
                   **type**: bool\n

           **Output**

           .. important::
                This method only prints the list of all models in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_models()
            >>> client.repository.list_models(limit=100)
            >>> client.repository.list_models(limit=100, get_all=True)
            >>> [entry for entry in client.repository.list_models(limit=100, asynchronous=True, get_all=True)]
        """
        return self._client._models.list(limit=limit, asynchronous=asynchronous, get_all=get_all)

    def list_experiments(self, limit=None):
        """
           List stored experiments. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all experiments in a table format.\n
                **return type**: None\n


           **Example**

            >>> client.repository.list_experiments()
        """
        if self._client.WSD:
            raise WMLClientError(u'Experiment APIs are not supported for Watson Studio Desktop.')
        self._client.experiments.list(limit=limit)

    def list_spaces(self, limit=None):
        """
           List stored spaces. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all spaces in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_spaces()
        """

        if self._client.WSD:
            raise WMLClientError('list_spaces - Listing spaces is not supported for Watson Studio Desktop.')

        if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
            raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

        self._client.spaces.list(limit=limit)

    def list_functions(self, limit=None):
        """

            List stored functions. If limit is set to None there will be only first 50 records shown.

            **Parameters**

            .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

            **Output**

            .. important::
                This method only prints the list of all functions in a table format.\n
                **return type**: None\n

            **Example**

             >>> client.respository.list_functions()
        """
        self._client._functions.list(limit=limit)

    def list_pipelines(self, limit=None):
        """
           List stored pipelines. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all pipelines in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_pipelines()
        """

        self._client.pipelines.list(limit=limit)
    def list_members(self, space_uid ,limit=None):
            """
               List stored members of a space. If limit is set to None there will be only first 50 records shown.

               **Parameters**

               .. important::
                    #. **limit**:  limit number of fetched records\n
                       **type**: int\n

               **Output**

               .. important::
                    This method only prints the list of all members associated with a space in a table format.\n
                    **return type**: None\n

               **Example**

                >>> client.spaces.list_members()
            """
            if self._client.WSD:
                raise WMLClientError('list_members - Listing members is not supported for Watson Studio Desktop.')

            if Repository.cloud_platform_spaces or Repository.icp_platform_spaces:
                raise WMLClientError(u"Not supported in this release. Use methods in 'client.spaces' instead")

            self._client.spaces.list_members(space_uid=space_uid,limit=limit)

    def _check_artifact_type(self, artifact_uid):
        artifact_uid = str_type_conv(artifact_uid)
        Repository._validate_type(artifact_uid, u'artifact_uid', STR_TYPE, True)

        def _artifact_exists(response):
            return (response is not None) and (u'status_code' in dir(response)) and (response.status_code == 200)

        pool = Pool(processes=4)
        #headers =

        isIcp=self._ICP

        if self._client.WSD:
            endpoint = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + artifact_uid
            response = requests.get(
                endpoint,
                params=self._client._params()
            )

           # requestsget_url, (endpoint, self._client._get_headers(), self._client._params(), True))
            response_get = _artifact_exists(response)

            artifact_type = artifact_uid.rsplit(".")[0]
            artifact_list = ['wml_model', 'wml_pipeline', 'wml_function']
            artifact_type_exists = {artifact.rsplit('_')[-1]: (response_get and artifact == artifact_type) for artifact in artifact_list}
            return artifact_type_exists

        else:
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                endpoints = {
                    u'model': self._client.service_instance._href_definitions.get_model_last_version_href(artifact_uid),
                    u'pipeline': self._client.service_instance._href_definitions.get_pipeline_href(artifact_uid),
                    u'experiment': self._client.service_instance._href_definitions.get_experiment_href(artifact_uid),
                    u'function': self._client.service_instance._href_definitions.get_function_href(artifact_uid)
                }
            else:
                endpoints = {
                        u'model': self._client.service_instance._href_definitions.get_model_last_version_href(artifact_uid),
                        u'pipeline': self._client.service_instance._href_definitions.get_pipeline_href(artifact_uid),
                        u'experiment': self._client.service_instance._href_definitions.get_experiment_href(artifact_uid),
                        u'function': self._client.service_instance._href_definitions.get_function_href(artifact_uid),
                        u'runtime': self._client.service_instance._href_definitions.get_runtime_href(artifact_uid),
                        u'library': self._client.service_instance._href_definitions.get_custom_library_href(artifact_uid),
                        u'space': self._client.service_instance._href_definitions.get_space_href(artifact_uid)
                    }
            future = {}
            for artifact in endpoints:
                if (artifact=="library" or artifact=="runtime" or artifact=="space"):
                    params = None
                else:
                    params = self._client._params()
                future[artifact] = pool.apply_async(get_url, (endpoints[artifact], self._client._get_headers(), params , isIcp))

            # future_no_space = {artifact: pool.apply_async(get_url, (endpoints[artifact], headers, None, isIcp)) for artifact in endpoints if (artifact=="library" or artifact=="runtime" or artifact=="space")}
            # future.update(future_no_space)
            response_get = {artifact: None for artifact in endpoints}

            for artifact in endpoints:
                    try:
                        response_get[artifact] = future[artifact].get(timeout=180)
                        self._logger.debug(u'Response({})[{}]: {}'.format(endpoints[artifact], response_get[artifact].status_code, response_get[artifact].text))

                    except Exception as e:
                        self._logger.debug(u'Error during checking artifact type: ' + str(e))

            pool.close()
            artifact_type = {artifact: _artifact_exists(response_get[artifact]) for artifact in response_get}

            return artifact_type


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_revision(self, artifact_uid):
        """
        Create revision for passed artifact_uid.

        **Parameters**

        .. important::

            #. **artifact_uid**: Unique id of stored model, experiment, function or pipelines.\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: Artifact new revision metadata.\n
            **return type**: dict\n

        **Example**

            >>> details = client.repository.create_revision(artifact_uid)
        """
        artifact_uid = str_type_conv(artifact_uid)
        Repository._validate_type(artifact_uid, u'artifact_uid', STR_TYPE, True)

        uid_type = self._check_artifact_type(artifact_uid)
        if uid_type[u'experiment'] is True:
            return self._client.experiments.create_revision(artifact_uid)
        if uid_type[u'pipeline'] is True:
            return self._client.pipelines.create_revision(artifact_uid)
        else:
           raise WMLClientError(u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))

        return details


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def _get_revision_details(self, artifact_uid):
        """
           Get metadata of stored artifacts revisions.

           :param artifact_uid:  unique id of stored model or experiment or function or pipelines (optional)
           :type artifact_uid: {str_type}

           :returns: stored artifacts metadata
           :rtype: dict

           A way you might use me is:

           >>> details = client.repository.get_revision_details(artifact_uid)

        """
        artifact_uid = str_type_conv(artifact_uid)
        Repository._validate_type(artifact_uid, u'artifact_uid', STR_TYPE, True)

        uid_type = self._check_artifact_type(artifact_uid)

        if uid_type[u'experiment'] is True:
            details = self._client.experiments.get_revision_details(artifact_uid)
        if uid_type[u'pipeline'] is True:
            details = self._client.pipelines.get_revisions(artifact_uid)

        else:
            raise WMLClientError(u'Getting artifact details failed. Artifact uid: \'{}\' not found.'.format(artifact_uid))
        return details

    def list_models_revisions(self, model_uid, limit=None):
        """
           List stored model revisions. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **model_uid**:  Uniquie Id of the model \n
                   **type**: str\n

                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all revisions of given model ID in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_models_revisions(model_uid)
        """

        self._client._models.list_revisions(model_uid, limit=limit)

    def list_pipelines_revisions(self, pipeline_uid, limit=None):
        """
           List stored pipeline revisions. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **model_uid**:  Uniquie Id of the pipeline \n
                   **type**: str\n
           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all revisions of given pipeline ID in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_pipelines_revisions(pipeline_uid)
        """
        self._client.pipelines.list_revisions(pipeline_uid, limit=limit)

    def list_functions_revisions(self, function_uid, limit=None):
        """
           List stored function revisions. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **function_uid**:  Uniquie Id of the function \n
                   **type**: str\n
           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all revisions of given function ID in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_functions_revisions(function_uid)
        """
        self._client._functions.list_revisions(function_uid, limit=limit)

    def list_experiments_revisions(self, experiment_uid, limit=None):
        """
           List stored experiment revisions. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **experiment_uid**:  Uniquie Id of the experiment \n
                   **type**: str\n

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all revisions of given experiment ID in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.repository.list_experiments_revisions(experiment_uid)
        """
        self._client.experiments.list_revisions(experiment_uid, limit=limit)

    def promote_model(self, model_id: str, source_project_id: str, target_space_id: str):
        """
           Promote model from project to space. Supported only for IBM Cloud Pak® for Data.

           :param model_id: stored model
           :type model_id: {str_type}

           :param source_project_id: stored model
           :type source_project_id: {str_type}

           :param target_space_id: stored model
           :type target_space_id: {str_type}

           :returns: promoted model id
           :rtype: str

           A way you might use me is:

           >>> promoted_model_id = client.models.promote(model_id, source_project_id, target_space_id)
        """
        return self._client._models.promote(model_id, source_project_id, target_space_id)





#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,PKG_EXTN_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import PkgExtnMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os,json

_DEFAULT_LIST_LENGTH = 50


class PkgExtn(WMLResource):
    """
    Store and manage your software Packages Extension specs.

    """
    ConfigurationMetaNames = PkgExtnMetaNames()
    """MetaNames for Package Extensions creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, pkg_extn_uid):
        """
            Get package extensions details.

            **Parameters**

            .. important::
                #. **pkg_extn_details**:  Metadata of the package extensions\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: pkg_extn UID\n
                **return type**: str

            **Example**

             >>> pkg_extn_details = client.pkg_extn.get_details(pkg_extn_uid)
        """
        PkgExtn._validate_type(pkg_extn_uid, u'pkg_extn_uid', STR_TYPE, True)

        response = requests.get(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_uid), params=self._client._params(),
                                headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(self._handle_response(200, u'get hw spec details', response))
        else:
            return self._handle_response(200, u'get hw spec details', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props, file_path):
        """
                Create a package extensions.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the pacakge extension. To see available meta names use:\n
                                    >>> client.package_extensions.ConfigurationMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the package extensions\n
                    **return type**: dict\n

                **Example**

                 >>> meta_props = {
                 >>>    client.package_extensions.ConfigurationMetaNames.NAME: "skl_pipeline_heart_problem_prediction",
                 >>>    client.package_extensions.ConfigurationMetaNames.DESCRIPTION: "description scikit-learn_0.20",
                 >>>    client.package_extensions.ConfigurationMetaNames.TYPE: "conda_yml",
                 >>> }

                **Example**

                 >>> pkg_extn_details = client.package_extensions.store(meta_props=meta_props,file_path="/path/to/file")

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        PkgExtn._validate_type(meta_props, u'meta_props', dict, True)
        pkg_extn_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        pkg_extn_meta_json = json.dumps(pkg_extn_meta)

        PkgExtn._validate_type(file_path, u'file_path', str, True)
        #Step1  : Create an asset
        print("Creating package extensions")
        href = self._client.service_instance._href_definitions.get_pkg_extns_href()

        creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=pkg_extn_meta_json)

        pkg_extn_details = self._handle_response(201, u'creating new package_extensions', creation_response)

        # Step2: upload pkg extension file to presigned url
        if creation_response.status_code == 201:
            pkg_extn_asset_id = pkg_extn_details["metadata"]["asset_id"]
            pkg_extn_presigned_url = pkg_extn_details["entity"]["package_extension"]["href"]

            FILE_SIZE_LIMIT = int(0.3 * 1073741824)

            def read_in_chunks(chunk_size=FILE_SIZE_LIMIT):
                with open(file_path, 'rb') as file_object:
                    while True:
                        data = file_object.read(chunk_size)
                        if not data:
                            break
                        yield data

            def send_multipart_request(file, url):
                content_path = os.path.abspath(file)
                content_size = os.stat(content_path).st_size

                index = 0
                headers = {}

                for chunk in read_in_chunks(FILE_SIZE_LIMIT):
                    offset = index + len(chunk)
                    headers['Content-Range'] = 'bytes %s-%s/%s' % (index, offset - 1, content_size)
                    index = offset

                    response = requests.put(
                        url,
                        files={"file": (file, chunk, 'application/octet-stream')}
                    )

                return response

            if os.stat(file_path).st_size == 0:
                raise WMLClientError('Package extension file cannot be empty')

            if not self._ICP:
                put_response = requests.put(
                    pkg_extn_presigned_url,
                    data=read_in_chunks()
                )
            else:
                put_response = send_multipart_request(file_path, self._wml_credentials['url'] + pkg_extn_presigned_url)

            if put_response.status_code == 201 or put_response.status_code == 200:
                # Step3: Mark the upload complete
                complete_response = requests.post(
                    self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_asset_id) + "/upload_complete",
                    headers=self._client._get_headers(),
                    params=self._client._params()
                )
                if complete_response.status_code == 204:
                    print("SUCCESS")
                    return self._get_required_element_from_response(pkg_extn_details)
                else:
                    #print(complete_response.text) # remove print later
                    self._delete(pkg_extn_asset_id)
                    raise WMLClientError("Failed while creating a package extensions " + complete_response.text)
            else:
                self._delete(pkg_extn_asset_id)
                raise WMLClientError("Failed while creating a package extensions " + put_response.text)
        else:
            raise WMLClientError("Failed while creating a package extensions " + creation_response.text)

    def list(self):
        """
           List package extensions.

           **Output**

           .. important::
                This method only prints the list of all package extensionss in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.package_extensions.list()

        """

        href = self._client.service_instance._href_definitions.get_pkg_extns_href()

        response = requests.get(href, params=self._client._params(), headers=self._client._get_headers())

        self._handle_response(200, u'list pkg_extn', response)
        asset_details = self._handle_response(200, u'list assets', response)["resources"]
        pkg_extn_values = [
            (m[u'metadata'][u'name'],
             m[u'metadata'][u'asset_id'],
             m[u'entity'][u'package_extension'][u'type'],
             m[u'metadata'][u'created_at']) for
            m in asset_details]

        self._list(pkg_extn_values, [u'NAME', u'ASSET_ID', u'TYPE', u'CREATED_AT'], None, _DEFAULT_LIST_LENGTH)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(pkg_extn_details):
        """
                Get Unique Id of package extensions. Deprecated!! use get_id(pkg_extn_details) instead

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the package extensions \n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of package extension \n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.package_extensions.get_uid(pkg_extn_details)

        """
        PkgExtn._validate_type(pkg_extn_details, u'pkg_extn_details', object, True)
        #print(pkg_extn_details)
        PkgExtn._validate_type_of_details(pkg_extn_details, PKG_EXTN_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(pkg_extn_details, u'pkg_extn_details',
                                                           [u'metadata', u'asset_id'])

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(pkg_extn_details):
        """
                Get Unique Id of package extensions.

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the package extensions \n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of package extension \n
                    **return type**: str\n

                **Example**


                 >>> asset_id = client.package_extensions.get_id(pkg_extn_details)

        """

        return PkgExtn.get_uid(pkg_extn_details)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid_by_name(self, pkg_extn_name):
        """
                Get UID of package extensions. Deprecated!! Use get_id_by_name(pkg_extn_name) instead

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the package extension\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of package extension \n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.package_extensions.get_uid_by_name(pkg_extn_name)
        """
        PkgExtn._validate_type(pkg_extn_name, u'pkg_extn_name', STR_TYPE, True)


        parameters = self._client._params()
        parameters.update(name=pkg_extn_name)

        response = requests.get(self._client.service_instance._href_definitions.get_pkg_extns_href(),
                                params=parameters,
                                headers=self._client._get_headers())

        if response.status_code == 200:
            total_values = self._handle_response(200, u'get pkg extn', response)["total_results"]
            if total_values != 0:
                pkg_extn_details = self._handle_response(200, u'get pkg extn', response)["resources"]
                return pkg_extn_details[0][u'metadata'][u'asset_id']
            else:
                return "Not Found"


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id_by_name(self, pkg_extn_name):
        """
                Get ID of package extensions.

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the package extension\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of package extension \n
                    **return type**: str\n

                **Example**


                 >>> asset_id = client.package_extensions.get_id_by_name(pkg_extn_name)
        """
        return PkgExtn.get_uid_by_name(self, pkg_extn_name)


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(pkg_extn_details):
        """
            Get url of stored package extensions.

           **Parameters**

           .. important::
                #. **asset_details**:  package extensions details\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of package extensions details\n
                **return type**: str\n

           **Example**

             >>> pkg_extn_details = client.package_extensions.get_details(pkg_extn_uid)
             >>> pkg_extn_href = client.package_extensions.get_href(pkg_extn_details)

        """
        PkgExtn._validate_type(pkg_extn_details, u'pkg_extn_details', object, True)
        PkgExtn._validate_type_of_details(pkg_extn_details, PKG_EXTN_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(pkg_extn_details,
                                                           u'pkg_extn_details',
                                                           [u'entity', u'package_extension', u'href'])


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, pkg_extn_uid):
        """
            Delete a package extension.

            **Parameters**

            .. important::
                #. **pkg_extn_uid**: Unique Id of package extension\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.package_extensions.delete(pkg_extn_uid)

        """
        PkgExtn._validate_type(pkg_extn_uid, u'pkg_extn_uid', STR_TYPE, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_uid), params=self._client._params(),
                                headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete pkg extn specification', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def _delete(self, pkg_extn_uid):
        PkgExtn._validate_type(pkg_extn_uid, u'pkg_extn_uid', STR_TYPE, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_uid), params=self._client._params(),
                                   headers=self._client._get_headers())

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'pkg_extn_response', dict)
        try:
            if self._client.default_space_id is not None:
                new_el = {'metadata': {'space_id': response_data['metadata']['space_id'],
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:

                    href = self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

                    new_el = {'metadata': {'project_id': response_data['metadata']['project_id'],
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                           },
                              'entity': response_data['entity']

                              }
                else:
                    new_el = {'metadata': {'project_id': response_data['metadata']['project_id'],
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                       },
                             'entity': response_data['entity']

                            }
            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'href' in response_data['metadata']:
                href_without_host = response_data['metadata']['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href': href_without_host})

            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data.text)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def download(self, pkg_extn_id, filename):
        """
            Download a package extension.

            **Parameters**

            .. important::
                 #. **pkg_extn_uid**:  The Unique Id of the package extension to be downloaded\n
                    **type**: str\n

                 #. **filename**:  filename to be used for the downloaded file\n
                    **type**: str\n

            **Output**

            .. important::
                 **returns**: Path to the downloaded package extension content\n
                 **return type**: str\n

            **Example**

             >>> client.assets.download(asset_uid,"sample_conda.yml/custom_library.zip")

         """

        PkgExtn._validate_type(pkg_extn_id, u'pkg_extn_id', STR_TYPE, True)

        if self._WSD:
            import urllib
            pkg_extn_response = requests.get(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_id),
                                          params=self._client._params(),
                                          headers=self._client._get_headers())

            pkg_extn_details = self._handle_response(200, u'get assets', pkg_extn_response)

            artifact_content_url = pkg_extn_details['entity']['package_extension']['href']

            r = requests.get(artifact_content_url, params=self._client._params(), headers=self._client._get_headers(),
                             stream=True)
            if r.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), r)

            downloaded_asset = r.content
            try:
                with open(filename, 'wb') as f:
                    f.write(downloaded_asset)
                print(u'Successfully saved asset content to file: \'{}\''.format(filename))
                return os.getcwd() + "/" + filename
            except IOError as e:
                raise WMLClientError(u'Saving asset with artifact_url: \'{}\' failed.'.format(filename), e)
        else:
            pkg_extn_response = requests.get(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_id),
                                      params=self._client._params(),
                                      headers=self._client._get_headers())

            pkg_extn_details = self._handle_response(200, u'get assets', pkg_extn_response)

            artifact_content_url = pkg_extn_details['entity']['package_extension']['href']

            if pkg_extn_response.status_code == 200:
                if not self._ICP:
                    # att_response = requests.get(self._wml_credentials["url"]+artifact_content_url)
                    att_response = requests.get(artifact_content_url)
                else:
                    att_response = requests.get(self._wml_credentials["url"]+artifact_content_url)

                if att_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading package extension"),
                                            att_response)

                downloaded_asset = att_response.content
                try:
                    with open(filename, 'wb') as f:
                        f.write(downloaded_asset)
                    print(u'Successfully saved package extension content to file: \'{}\''.format(filename))
                    return os.getcwd() + "/" + filename
                except IOError as e:
                    raise WMLClientError(u'Saving asset with artifact_url: \'{}\' failed.'.format(filename), e)
            else:
                raise WMLClientError("Failed while downloading the package extension "+ pkg_extn_id)


__all__ = [
    "MissingPipeline",
    "FitNotCompleted",
    "FitNeeded",
    "AutoAIComputeError",
    "MissingAutoPipelinesParameters",
    "UseWMLClient",
    "PipelineNotLoaded",
    "MissingCOSStudioConnection",
    "MissingProjectLib",
    "LocalInstanceButRemoteParameter",
    "DataFormatNotSupported",
    "HoldoutSplitNotSupported",
    'LibraryNotCompatible',
    'CannotInstallLibrary',
    'InvalidCOSCredentials',
    'CannotDownloadTrainingDetails',
    'TShirtSizeNotSupported',
    'MissingPositiveLabel',
    'MissingDataPreprocessingStep',
    'CannotDownloadWMLPipelineDetails',
    'WrongDataJoinGraphNodeName',
    'NotInWatsonStudio',
    'WrongWMLServer',
    'CredentialsNotFound',
    'SetIDFailed',
    'MissingLocalAsset',
    'DataSourceSizeNotSupported',
    'TrainingDataSourceIsNotFile',
    'VisualizationFailed',
    'InvalidPredictionType',
    'InvalidIdType',
    'NoneDataConnection',
    'CannotReadSavedRemoteDataBeforeFit',
    'NotWSDEnvironment',
    'NotExistingCOSResource',
    'AdditionalParameterIsUnexpected',
    'OBMForNFSIsNotSupported',
    'ForecastPredictionColumnsMissing',
    'NonForecastPredictionColumnMissing',
    'ForecastingCannotBeRunAsLocalScenario',
    'ForecastingUnsupportedOperation',
    'OBMMainNodeAlreadySet',
    'InvalidSequenceValue',
    'NoAvailableMetrics',
    'WrongAssetType',
    'TSNotSupported',
    'InvalidDataAsset',
    'ParamOutOfRange',
    'CannotConnectToWebsocket',
    'TestDataNotPresent',
    'NoAutomatedHoldoutSplit',
    'DiscardedModel',
    'WrongModelName',
    'StrategyIsNotApplicable',
    'NumericalImputationStrategyValueMisused',
    'InvalidImputationParameterNonTS',
    'InvalidImputationParameterTS',
    'InconsistentImputationListElements',
    'ImputationListNotSupported',
    'MissingEstimatorForExistingBatchedEstimator',
    'FutureExogenousFeaturesNotSupported',
    'NoAvailableNotebookLocation',
    "ContainerTypeNotSupported"
]

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibm_watson_machine_learning.utils import WMLClientError
from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, ImputationStrategy


class MissingPipeline(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"There is no such a Pipeline like: {value_name}", reason)


class FitNotCompleted(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Fit run is not completed or the status is failed for run: {value_name}", reason)


class FitNeeded(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Fit run was not performed.", reason)


class AutoAIComputeError(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Fit run failed for run_id: {value_name}.", reason)


class MissingAutoPipelinesParameters(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"AutoPipelines parameters are {value_name}", reason)


class UseWMLClient(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Use WML v4 Client instead of {value_name}", reason)


class PipelineNotLoaded(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Pipeline model: {value_name} cannot load.", reason)


class MissingCOSStudioConnection(WMLClientError, ValueError):
    def __init__(self, reason=None):
        WMLClientError.__init__(self, f"Missing COS Studio connection.", reason)


class MissingProjectLib(WMLClientError, ValueError):
    def __init__(self, reason=None):
        WMLClientError.__init__(self,
                                f"project-lib package missing in the environment, please make sure you are on Watson Studio"
                                f" and want to automatically initialize your COS connection."
                                f" If you want to initialize COS connection manually, do not use from_studio() method.",
                                reason)


class LocalInstanceButRemoteParameter(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Provided {value_name} parameter to local optimizer instance.", reason)


class DataFormatNotSupported(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self,
                                f"This data format is not supported by SDK. (CSV and XLSX formats are supported.)",
                                reason)


class HoldoutSplitNotSupported(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Holdout split is not supported for xlsx data.", reason)


class LibraryNotCompatible(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Library not compatible or missing!", reason)


class CannotInstallLibrary(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Library cannot be installed! Error: {value_name}", reason)


class InvalidCOSCredentials(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Wrong COS credentials!", reason)


class CannotDownloadTrainingDetails(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Cannot download training details, training is not done yet. "
                                      f"Please try again after training is finished.", reason)


class TShirtSizeNotSupported(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"This t-shirt size: \"{value_name}\" is not supported on this environment.",
                                reason)


class MissingPositiveLabel(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Missing positive label for \"{value_name}\"", reason)


class MissingDataPreprocessingStep(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Data preprocessing step not performed.", reason)


class CannotDownloadWMLPipelineDetails(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Cannot download WML pipeline details ", reason)


class WrongDataJoinGraphNodeName(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"{value_name} is not a correct node name.", reason)


class NotInWatsonStudio(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Wrong environment.", reason)


class WrongWMLServer(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Wrong WML Server instance.", reason)


class CredentialsNotFound(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Cannot find WMLS credentials in WSD.", reason)


class SetIDFailed(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Cannot set {value_name}.", reason)


class MissingLocalAsset(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Local asset: {value_name} cannot be found.", reason)


class DataSourceSizeNotSupported(WMLClientError, ValueError):
    def __init__(self, reason=None):
        WMLClientError.__init__(self, f"The selected data source is too large for selected compute configuration "
                                      f"and might fail to run. Consider increasing the compute configuration", reason)


class TrainingDataSourceIsNotFile(WMLClientError, ValueError):
    def __init__(self, data_location=None, reason=None):
        WMLClientError.__init__(self, f"Training data location: {data_location} is a directory or does not exist."
                                      f"Please set training data location to dataset file location.", reason)


class VisualizationFailed(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Cannot perform visualization.", reason)


class InvalidPredictionType(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Cannot recognize prediction type: {value_name}", reason)


class InvalidIdType(WMLClientError, ValueError):
    def __init__(self, typ):
        WMLClientError.__init__(self, f"Unexpected type of id: {typ}")


class NoneDataConnection(WMLClientError, ValueError):
    def __init__(self, resource_name):
        WMLClientError.__init__(self, f"Invalid DataConnection in {resource_name} list. DataConnection cannot be None")


class CannotReadSavedRemoteDataBeforeFit(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, f"Cannot read saved remote data before fit.")


class NotWSDEnvironment(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, f"Function cannot be used outside WSD environment.")


class NotExistingCOSResource(WMLClientError, ValueError):
    def __init__(self, bucket, path):
        WMLClientError.__init__(self, f"There is no COS resource: {path} in bucket: {bucket}.")


class AdditionalParameterIsUnexpected(WMLClientError):
    def __init__(self, param):
        WMLClientError.__init__(self, f"Additional parameter is not expected: {param}")


class OBMForNFSIsNotSupported(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self, f"OBM for NFS is not supported.")


class ForecastPredictionColumnsMissing(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self,
                                "For prediction_type = `forecasting` passing `prediction_columns` is required, while `prediction_column` should be None.")


class NonForecastPredictionColumnMissing(WMLClientError):
    def __init__(self, prediction_type):
        WMLClientError.__init__(self,
                                f"For prediction_type = `{prediction_type}` passing `prediction_column` is required, while `prediction_columns` should be None.")


class ForecastingCannotBeRunAsLocalScenario(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self, f"Forecasting cannot be run in local scenario.")


class ForecastingUnsupportedOperation(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self, f"Operation is unsupported for timeseries forecasting prediction type.")


class OBMMainNodeAlreadySet(WMLClientError):
    def __init__(self, param):
        WMLClientError.__init__(self, f"OBM Main node is already set to: {param}. If you want to set the new one,"
                                      f"please reinitialize entire object.")


class InvalidSequenceValue(WMLClientError, ValueError):
    def __init__(self, el, correct_values):
        WMLClientError.__init__(self, f"Invalid sequence element: '{el}' sequence must be composed from "
                                      f"given values: {correct_values}")


class NoAvailableMetrics(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self, f"Currently there is no available metrics.")


class WrongAssetType(WMLClientError):
    def __init__(self, asset_type: str):
        WMLClientError.__init__(self, f"This asset type: '{asset_type}' is not supported.")


class TSNotSupported(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self, f"Time series forecasting is not supported for CPD 2.5, 3.0 and 3.5.")


class InvalidDataAsset(WMLClientError, ValueError):
    def __init__(self, reason='Wrong Data Asset'):
        WMLClientError.__init__(self, "Data Asset without 'connection' is not supported for result store action.",
                                reason)


class ParamOutOfRange(WMLClientError, ValueError):
    def __init__(self, param_name, value, min, max):
        WMLClientError.__init__(self,
                                f"Value of parameter `{param_name}`, {value}, is out of expected range - between {min} and {max}.")


class CannotConnectToWebsocket(WMLClientError, ValueError):
    def __init__(self, n):
        WMLClientError.__init__(self, f"{n} times connecting to websocket failed.")


class TestDataNotPresent(WMLClientError, ValueError):
    def __init__(self, reason):
        WMLClientError.__init__(self, "User defined (test / holdout) data is not present for this AutoAI experiment.",
                                reason)


class NoAutomatedHoldoutSplit(WMLClientError, ValueError):
    def __init__(self, reason):
        WMLClientError.__init__(self, "AutoAI experiment was performed with user defined holdout."
                                      "To recreate holdout split, please use test data connection."
                                      "You can call optimizer.get_test_data_connections().",
                                reason)


class DiscardedModel(WMLClientError, ValueError):
    def __init__(self, model_name):
        WMLClientError.__init__(self,
                                "You are trying to store a discarded forecasting model pipeline. Please use not discarded one."
                                "Look at the pipelines 'summary' method and choose the 'Winner' one.",
                                reason=f"Pipeline: '{model_name}' is discarded!")


class WrongModelName(WMLClientError, ValueError):
    def __init__(self, model_name):
        WMLClientError.__init__(self,
                                "This model name does not exist. Please provide correct one."
                                "Look at the pipelines 'summary' method and choose the 'Winner' one.",
                                reason=f"Pipeline: '{model_name}' does not exist!")


class StrategyIsNotApplicable(WMLClientError, ValueError):
    def __init__(self, strategy, prediction_type, valid_strategies):
        WMLClientError.__init__(self,
                                f'{strategy} is not valid for '
                                f'{"forecasting" if prediction_type == PredictionType.FORECASTING else "non-forecasting"} '
                                f'scenario. Valid strategies are: {", ".join(valid_strategies)}')


class NumericalImputationStrategyValueMisused(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, 'Parameter `numerical_imputation_value` can only be used when '
                                      '`numerical_imputation_strategy` is set to ImputationStrategy.VALUE.')


class InvalidImputationParameterNonTS(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, '`numerical_imputation_value` and `imputation_threshold` can be set only for forecasting experiments.')


class InvalidImputationParameterTS(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, '`categorical_imputation_strategy` does not apply to a forecasting experiment.')


class InconsistentImputationListElements(WMLClientError, ValueError):
    def __init__(self, strategies):
        l = ['ImputationStrategy.' + s.name for s in strategies]
        WMLClientError.__init__(self, f'`categorical_imputation_strategy` list elements are not compatible: {", ".join(l)}')


class ImputationListNotSupported(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, f'List passed as imputation strategy is not supported for non-forecasting scenarios.')


class InvalidSamplingType(WMLClientError, ValueError):
    def __init__(self, sampling_type: str, predcition_type: str):
        WMLClientError.__init__(self, f"Sampling type for {sampling_type} is invalid for prediction type {predcition_type}.")


class MissingEstimatorForExistingBatchedEstimator(WMLClientError, ValueError):
    def __init__(self, batched_estimator: str, missing_estimator: str):
        WMLClientError.__init__(self, f"There is no corresponding estimator in `include_only_estimators` list for " +
                                f"{batched_estimator} estimator in `include_batched_ensemble_estimators` list. " +
                                f"Add {missing_estimator} estimator to `include_only_estimators` list.")


class FutureExogenousFeaturesNotSupported(WMLClientError, ValueError):
    def __init__(self, prediction_type: str):
        WMLClientError.__init__(self, f"Future Exogenous features are only supported for forecasting."
                                      f"Current prediction type is: {prediction_type}")


class NoAvailableNotebookLocation(WMLClientError, ValueError):
    def __init__(self, pipeline_name):
        WMLClientError.__init__(self, f"Unable to find notebook location in training details for {pipeline_name}. " +
                                "Possible causes: pipeline failure or pipeline not classified as winner in " +
                                "forecasting scenario.")


class ContainerTypeNotSupported(WMLClientError, ValueError):
    def __init__(self):
        WMLClientError.__init__(self, f"The container data connection type is not supported for CP4D environment."
                                      f" Supported types are: data asset and connection asset. ")


#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from  ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.scikit_artifact_loader import ScikitArtifactLoader


class ScikitPipelineModelLoader(ScikitArtifactLoader):
    """
        Returns pipeline model instance associated with this model artifact.

        :return: pipeline model
        :rtype: scikit.learn.Pipeline
        """
    def model_instance(self,as_type=None):
        """
           :param as_type: string type referring to the model type to be returned.
           This parameter is applicable for xgboost models only.
           Currently accepts:
             'Booster': returns a model of type xgboost.Booster
             'XGBRegressor': returns a model of type xgboost.sklearn.XGBRegressor
           :return: returns a scikit model or an xgboost model of type xgboost.Booster
           or xgboost.sklearn.XGBRegressor
         """
        return self.load(as_type)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark.ml.common import inherit_doc
from pyspark.ml.util import MLReadable, JavaMLReader, _jvm


class PipelineJavaMLReader(JavaMLReader):
    """
    MLPipeline JavaML Reader
    """

    def __init__(self, owner_clazz, class_name):
        self._className = class_name
        super(PipelineJavaMLReader, self).__init__(owner_clazz)

    def _java_loader_class(self, clazz):
        return self._className

    def _load_java_obj(self, clazz):
        """Load the peer Java object of the ML instance."""
        java_class = self._java_loader_class(clazz)
        java_obj = _jvm()
        for name in java_class.split("."):
            java_obj = getattr(java_obj, name)
        return java_obj








#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE, \
    DATA_ASSETS_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, \
    get_file_from_cos
from ibm_watson_machine_learning.metanames import ConnectionMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os
import json
from urllib.parse import unquote, quote

_DEFAULT_LIST_LENGTH = 0


class Connections(WMLResource):
    """
    Store and manage your Connections.
    """
    ConfigurationMetaNames = ConnectionMetaNames()
    """MetaNames for Connection creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'connection_response', dict)
        try:
            new_el = {'metadata': {'id': response_data['metadata']['asset_id'],
                                   'asset_type': response_data['metadata']['asset_type'],
                                   'create_time': response_data['metadata']['create_time'],
                                   'last_access_time': response_data['metadata']['usage'].get('last_access_time')
                                   },
                      'entity': {
                          'datasource_type': response_data['entity']['datasource_type'],
                          'name': response_data['entity']['name']
                      }
                      }

            for el in ['description', 'origin_country', 'owner_id', 'properties']:
                if el in response_data['entity']:
                    new_el['entity'][el] = response_data['entity'].get(el)

            if self._client.default_space_id is not None:
                new_el['metadata']['space_id'] = response_data['metadata']['space_id']

            elif self._client.default_project_id is not None:
                new_el['metadata']['project_id'] = response_data['metadata']['project_id']

                if not self._client.WSD:
                    if 'href' in response_data['metadata']:
                        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                            href_without_host = response_data['href'].split('.com')[-1]
                            new_el[u'metadata'].update({'href': href_without_host})
                        else:
                            new_el['metadata'].update({'href': response_data['href']})

            return new_el
        except Exception as e:
            raise WMLClientError(Messages.get_message(response_data, message_id="failed_to_read_response_from_down_stream_service"))

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, connection_id=None):
        """
        Get connection details for the given unique Connection id. If no connection_id is passed, details for all connections will be returned.

        **Parameters**

        .. important::
            #. **connection_id**: Unique id of Connection\n
                **type**: str\n

        **Output**

        .. important::
            **returns**: Metadata of the stored Connection\n
            **return type**: dict\n

        **Example**

         >>> connection_details = client.connections.get_details(connection_id)
         >>> connection_details = client.connections.get_details()
        """
        self._client._check_if_either_is_set()
        Connections._validate_type(connection_id, u'connection_id', STR_TYPE, False)

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        if self._client._iam_id:
            header_param['IBM-WDP-Impersonate'] = str({'iam_id': str(self._client._iam_id)})

        if connection_id:
            with self.requests_retry_session() as sess:
                response = sess.get(
                    self._client.service_instance._href_definitions.get_connection_by_id_href(connection_id),
                    params=self._client._params(),
                    headers=header_param)

                if response.status_code == 200:
                    return self._get_required_element_from_response(
                        self._handle_response(200, u'get connection details', response))
                else:
                    return self._handle_response(200, u'get connection details', response)
        else:
            with self.requests_retry_session() as sess:
                response = sess.get(self._client.service_instance._href_definitions.get_connections_href(),
                                    params=self._client._params(),
                                    headers=header_param)

                if response.status_code == 200:
                    return {'resources': [self._get_required_element_from_response(x) for x in response.json()['resources']]}
                else:
                    return self._handle_response(200, u'get connection details', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create(self, meta_props):
        """
                Creates a connection. Input to PROPERTIES field examples

                1. MySQL
                      >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                      >>>   "database": "database",
                      >>>   "password": "password",
                      >>>   "port": "3306",
                      >>>   "host": "host url",
                      >>>   "ssl": "false",
                      >>>   "username": "username"
                      >>> }

                2. Google Big query

                   a. Method1: Use service account json. The service account json generated can be provided as
                               input as-is. Provide actual values in json. Example is only indicative to show
                               the fields. Refer to Google big query documents how to generate the service account json\n

                      >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                      >>>     "type": "service_account",
                      >>>     "project_id": "project_id",
                      >>>     "private_key_id": "private_key_id",
                      >>>     "private_key": "private key contents",
                      >>>     "client_email": "client_email",
                      >>>     "client_id": "client_id",
                      >>>     "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                      >>>     "token_uri": "https://oauth2.googleapis.com/token",
                      >>>     "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                      >>>     "client_x509_cert_url": "client_x509_cert_url"
                      >>> }

                   b. Method2: Using OAuth Method. Refer to Google big query documents how to generate OAuth token\n

                       >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                       >>>    "access_token": "access token generated for big query",
                       >>>    "refresh_token": "refresh token",
                       >>>    "project_id": "project_id",
                       >>>    "client_secret": "This is your gmail account password",
                       >>>   "client_id": "client_id"
                       >>> }

                3. MS SQL
                     >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                     >>>      "database": "database",
                     >>>      "password": "password",
                     >>>      "port": "1433",
                     >>>      "host": "host",
                     >>>      "username": "username"
                     >>> }

                4. Tera data
                     >>> client.connections.ConfigurationMetaNames.PROPERTIES: {
                     >>>      "database": "database",
                     >>>      "password": "password",
                     >>>      "port": "1433",
                     >>>      "host": "host",
                     >>>      "username": "username"
                     >>> }


                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the connection configuration. To see available meta names use:\n
                            >>> client.connections.ConfigurationMetaNames.get()

                      **type**: dict\n
                **Output**

                .. important::

                    **returns**: metadata of the stored connection\n
                    **return type**: dict\n

                **Example**

                 >>> sqlserver_data_source_type_id = client.connections.get_datasource_type_uid_by_name('sqlserver')
                 >>> connections_details = client.connections.create({
                 >>> client.connections.ConfigurationMetaNames.NAME: "sqlserver connection",
                 >>> client.connections.ConfigurationMetaNames.DESCRIPTION: "connection description",
                 >>> client.connections.ConfigurationMetaNames.DATASOURCE_TYPE: sqlserver_data_source_type_id,
                 >>> client.connections.ConfigurationMetaNames.PROPERTIES: { "database": "database",
                 >>>                                                         "password": "password",
                 >>>                                                         "port": "1433",
                 >>>                                                         "host": "host",
                 >>>                                                         "username": "username"}
                 >>> })

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        connection_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        big_query_data_source_type_id = self.get_datasource_type_uid_by_name('bigquery')

        # Either service acct json credentials can be given or oauth json can be given
        # If service acct json, then we need to create a newline json with "credentials" key
        if connection_meta[u'datasource_type'] == big_query_data_source_type_id:
            if 'private_key' in connection_meta[u'properties']:
                result = json.dumps(connection_meta[u'properties'], separators=(',\n', ':'))
                newmap = {"credentials": result}
                connection_meta[u'properties'] = newmap

        connection_meta.update({'origin_country': 'US'})
        #Step1  : Create an asset
        print(Messages.get_message(message_id="creating_connections"))

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        creation_response = requests.post(
            self._client.service_instance._href_definitions.get_connections_href(),
            headers=header_param,
            json=connection_meta,
            params=self._client._params()
        )
        connection_details = self._handle_response(201, u'creating new connection', creation_response)
        if creation_response.status_code == 201:
            connection_id = connection_details["metadata"]["asset_id"]
            print(Messages.get_message(message_id="success"))
            return self._get_required_element_from_response(connection_details)
        else:
            raise WMLClientError(Messages.get_message(message_id="failed_while_creating_connections"))

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, connection_id):
        """
            Delete a stored Connection.

            **Parameters**

            .. important::
                #. **connection_id**: Unique id of the connection to be deleted.\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.connections.delete(connection_id)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        pipeline_uid = str_type_conv(connection_id)
        Connections._validate_type(connection_id, u'connection_id', STR_TYPE, True)
        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        connection_endpoint = self._client.service_instance._href_definitions.get_connection_by_id_href(connection_id)
        response_delete = requests.delete(connection_endpoint, params=self._client._params(),
                                          headers=header_param)

        return self._handle_response(204, u'connection deletion', response_delete, False)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(connection_details):
        """
                Get Unique Id of stored connection.

                **Parameters**

                .. important::

                   #. **connection_details**:  Metadata of the stored connection\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored connection\n
                    **return type**: str\n

                **Example**


                 >>> connection_uid = client.connection.get_uid(connection_details)

        """
        Connections._validate_type(connection_details, u'connection_details', object, True)

        return WMLResource._get_required_element_from_dict(connection_details, u'connection_details',
                                                           [u'metadata', u'id'])

    def list_datasource_types(self):
        """
        List stored datasource types assets.


        **Output**

        .. important::
            This method only prints the list of datasources type in a table format.\n
            **return type**: None\n

        **Example**

         >>> client.connections.list_datasource_types()
        """

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        response = requests.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                headers=header_param)

        datasource_details = self._handle_response(200, u'list datasource types', response)['resources']
        space_values = [
            (m[u'entity'][u'name'], m[u'metadata'][u'asset_id'], m[u'entity'][u'type'], m['entity']['status']) for
            m in datasource_details]

        self._list(space_values, [u'NAME', u'DATASOURCE_ID', u'TYPE', u'STATUS'], None, None)

    def list(self):
        """
           List all stored connections.

           **Output**

           .. important::
                This method only prints the list of all connections in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.connections.list()

        """
        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        response = requests.get(self._client.service_instance._href_definitions.get_connections_href(),
                                params=self._client._params(),
                                headers=header_param)

        self._handle_response(200, u'list datasource type', response)

        datasource_details = self._handle_response(200, u'list datasource types', response)['resources']
        space_values = [
            (m[u'entity'][u'name'], m[u'metadata'][u'asset_id'], m['metadata']['create_time'],
             m[u'entity'][u'datasource_type']) for
            m in datasource_details]

        self._list(space_values, [u'NAME', u'ID', u'CREATED', u'DATASOURCE_TYPE_ID', ], None, None)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def list_uploaded_db_drivers(self):
        """
           List uploaded db driver jars.
           Supported for IBM Cloud Pak for Data only.

           **Output**

           .. important::
                This method only prints the uploaded db driver jar names.\n
                **return type**: Str\n

           **Example**

            >>> client.connections.list_uploaded_db_drivers()

        """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')
        
        response = requests.get(
            self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + 'dbdrivers',
            headers=self._client._get_headers(no_content_type=True),
            params=self._client._params(),
            verify=not self._client.ICP
        )
        jars = [[el['path'].split('/')[-1]] for el in response.json()['resources']]

        self._list(jars, [u'NAME'], None, None)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_datasource_type_uid_by_name(self, name):
        """
           Get stored datasource types id for the given datasource type name.

           **Parameters**

           .. important::
                #. **name**:  name of datasource type\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the id of given datasource type name.\n
                **return type**: Str\n

           **Example**

            >>> client.connections.get_datasource_type_uid_by_name('cloudobjectstorage')

        """
        Connections._validate_type(name, u'name', str, True)

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        with self.requests_retry_session() as sess:
            response = sess.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                headers=header_param)

            datasource_id = 'None'
            datasource_details = self._handle_response(200, u'list datasource types', response)['resources']

        for i, ds_resource in enumerate(datasource_details):
            if ds_resource['entity']['name'] == name:
                datasource_id = ds_resource['metadata']['asset_id']
        return datasource_id

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def upload_db_driver(self, path: str):
        """
           Upload db driver jar.
           Supported for IBM Cloud Pak for Data only, version 4.0.4 and above.

           **Parameters**

           .. important::
                #. **path**:  path to db driver jar\n
                   **type**: str\n

           **Example**

            >>> client.connections.upload_db_driver('example/path/db2jcc4.jar')

        """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        driver_file_name = path.split('/')[-1]

        with open(path, 'rb') as fdata:
            content_upload_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                 "dbdrivers/" + quote(driver_file_name, safe='')
            response = requests.put(
                content_upload_url,
                files={'file': ('native', fdata, 'application/octet-stream', {'Expires': '0'})},
                headers=self._client._get_headers(no_content_type=True),
                params=self._client._params()
            )

            self._client.repository._handle_response(201, 'uploading db driver jar', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def sign_db_driver_url(self, jar_name: str):
        """
           Get signed db driver jar url to be used during creating of JDBC generic connection. The jar name passed as argument needs to be uploaded into system first.
           Supported for IBM Cloud Pak for Data only, version 4.0.4 and above..

           **Parameters**

           .. important::
                #. **jar_name**:  db driver jar name\n
                   **type**: str\n

           **Example**

            >>> client.connections.sign_db_driver_url('db2jcc4.jar')

        """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        signed_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                     quote("dbdrivers/" + jar_name, safe='') + "/signed"
        params = self._client._params()

        params['expires_in'] = 5000

        response = requests.post(
            signed_url,
            headers=self._client._get_headers(no_content_type=True),
            params=params
        )

        self._client.repository._handle_response(201, 'signing db driver url', response, json_response=False)

        return unquote(response.headers['Location'])

    @staticmethod
    def requests_retry_session(retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 503, 504, 521, 524), session=None):
        from requests.adapters import HTTPAdapter
        from requests.packages.urllib3.util.retry import Retry

        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            raise_on_status=False
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pyspark import SparkContext
from py4j.java_collections import MapConverter
from pyspark.sql import DataFrame,SQLContext
import inspect

class Source (object):
    """
    Class for creating a Source object to fetch data from an external DataSource like AmazonS3, Swift, DashDB etc.
    A Source object can be used as the first stage of a :class:`pipeline.DAG.DAG`

    :param symbolicconst: Connection string to access the external DataSource.
    :param optionsmap: A dictionary specifying the options to be passed to the DataSource.
    """
    _sc = SparkContext._active_spark_context
    _logSrcLang = "Py:"

    def __init__(self, symbolicconst, optionsmap):

        self._jPipeline = self._sc._jvm.com.ibm.analytics.ngp.pipeline.pythonbinding.Pipelines
        self._jLogger = self._sc._jvm.org.apache.log4j.Logger
        self._to_map = self._jPipeline.toScalaMap
        self.joptionsmap = MapConverter().convert(optionsmap, self._sc._gateway._gateway_client)
        sca_map = self._to_map(self.joptionsmap)
        self.logger = self._jLogger.getLogger(self.__class__.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: symbolicconst => " + str(
            symbolicconst) + " | optionsmap => " + str(optionsmap) + "]"
        self.logger.info(logMsg)

        self._jSource = self._jPipeline.getSource().apply(symbolicconst,sca_map)

    def run(self):
        """
        Used to retrieve the DataFrame from a :class:`Source` object that refers to an external DataSource.

        :return: A DataFrame
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: None]"
        self.logger.info(logMsg)

        return DataFrame(self._jSource.run(), SQLContext.getOrCreate(self._sc))

    def _to_java(self):
        return self._jSource



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import joblib
import numpy as np
from sklearn.cluster import KMeans
from sklearn.exceptions import NotFittedError

from ibmfl.util import config
from ibmfl.model.sklearn_fl_model import SklearnFLModel
from ibmfl.model.model_update import ModelUpdate
from ibmfl.exceptions import LocalTrainingException, \
    ModelInitializationException, ModelException

logger = logging.getLogger(__name__)


class SklearnKMeansFLModel(SklearnFLModel):
    """
    Wrapper class for sklearn.cluster.KMeans
    """

    def __init__(self, model_name, model_spec, sklearn_model=None, **kwargs):
        """
        Create a `SklearnKMeansFLModel` instance from a
        sklearn.cluster.KMeans model.
        If sklearn_model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: A name specifying the type of model, e.g., \
        clustering_KMeans
        :type model_name: `str`
        :param model_spec: A dictionary contains model specification
        :type model_spec: `dict`
        :param sklearn_model: Complied sklearn model
        :type sklearn_model: `sklearn.cluster.KMeans`
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a sklearn KMeans model.
        :type kwargs: `dict`
        """
        super().__init__(model_name, model_spec,
                         sklearn_model=sklearn_model,
                         **kwargs)
        self.model_type = 'Sklearn-KMeans'
        if sklearn_model:
            if not issubclass(type(sklearn_model), KMeans):
                raise ValueError('Compiled sklearn model needs to be provided'
                                 '(sklearn.cluster.KMeans). '
                                 'Type provided: ' + str(type(sklearn_model)))

            self.model = sklearn_model

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Tuple with first elements being the training data \
        (x_train,)
        :type train_data: `np.ndarray`
        :param fit_params: (optional) Dictionary with hyperparameters that \
        will be used to call sklearn.cluster fit function. \
        Provided hyperparameter should only contains parameters that \
        match sklearn expected values, e.g., `n_clusters`, which provides \
        the number of clusters to fit. \
        If no `fit_params` is provided, default values as defined in sklearn \
        definition are used.
        :return: None
        """

        # Extract x_train by default,
        # Only x_train is extracted since Clustering is unsupervised

        x_train = train_data[0]
        
        hyperparams = fit_params.get('hyperparams', {}) or {} if fit_params else {}
        local_hp = hyperparams.get('local', {}) or {}
        training_hp = local_hp.get('training', {}) or {}

        try:
            self.model.set_params(**training_hp)
        except Exception as err:
            logger.exception(str(err))
            raise LocalTrainingException(
                'Error occurred while setting up model parameters')

        try:
            self.model.fit(x_train)
        except Exception as err:
            logger.info(str(err))
            raise LocalTrainingException(
                'Error occurred while performing model.fit'
            )

    def update_model(self, model_update):
        """
        Update sklearn model with provided model_update, where model_update
        should contain `cluster_centers_` having the same
        dimension as expected by the sklearn.cluster model.
        `cluster_centers_` : np.ndarray, shape (n_clusters, n_features)

        :param model_update: `ModelUpdate` object that contains the \
        cluster_centers vectors that will be used to update the model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if isinstance(model_update, ModelUpdate):
            cluster_centers_ = model_update.get('weights')

            try:
                if cluster_centers_ is not None:
                    self.model.cluster_centers_ = np.array(cluster_centers_)
            except Exception as err:
                raise LocalTrainingException('Error occurred during '
                                             'updating the model weights. ' +
                                             str(err))
        else:
            raise LocalTrainingException('Provided model_update should be of '
                                         'type ModelUpdate. '
                                         'Instead they are: ' +
                                         str(type(model_update)))

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        try:
            cluster_centers_ = self.model.cluster_centers_
        except AttributeError:
            cluster_centers_ = None

        return ModelUpdate(weights=cluster_centers_)

    def evaluate(self, test_dataset, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, test) or a datagenerator of of type `keras.utils.Sequence`, 
        `keras.preprocessing.image.ImageDataGenerator`
        :type test_dataset: `np.ndarray`

        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test)

        else:
            raise ModelException("Invalid test dataset!")

    def evaluate_model(self, x, y, **kwargs):
        """
        Evaluates the model given test data x and the corresponding labels y.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Not used for evaluation since this is an unsupervised model
        :type y: `None`
        :param kwargs: Dictionary of model-specific arguments \
        for evaluating models. For example, sample weights accepted \
        by model.score.
        :return: Dictionary with all evaluation metrics provided by \
        specific implementation.
        :rtype: `dict`
        """
        acc = {}

        try:
            acc['score'] = self.model.score(x, **kwargs)
        except NotFittedError:
            logger.info('Model evaluated before fitted. '
                        'Returning accuracy as 0')
            acc['score'] = 0

        return acc

    def save_model(self, filename=None, path=None):
        """
        Save a sklearn.cluster.KMeans model to file in the format specific
        to the framework requirement. Meanwhile, initialize the attribute
        `_n_threads` to 1 if the model does not have it.

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path \
        is specified, the model will be stored in the default data location of \
        the library `DATA_PATH`.
        :type path: `str`
        :return: filename
        """
        if not hasattr(self.model, '_n_threads'):
            logger.info("Attribute _n_threads does not exist. "
                        "Setting it to default value.")
            self.model._n_threads = 1

        return super().save_model(filename, path)

    @staticmethod
    def load_model_from_spec(model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains the following items: model_spec['model_definition']
        contains the model definition as type sklearn.cluster.KMeans.

        :param model_spec: Model specification contains \
        a compiled sklearn model.
        :type model_spec: `dict`
        :return: model
        :rtype: `sklearn.cluster`
        """
        model = None
        try:
            if 'model_definition' in model_spec:
                model_file = model_spec['model_definition']
                model_absolute_path = config.get_absolute_path(model_file)

                with open(model_absolute_path, 'rb') as f:
                    model = joblib.load(f)

                if not issubclass(type(model), KMeans):
                    raise ValueError('Provided complied model in model_spec '
                                     'should be of type sklearn.cluster.'
                                     'Instead they are: ' + str(type(model)))
        except Exception as ex:
            raise ModelInitializationException('Model specification was '
                                               'badly formed. '+ str(ex))
        return model
